"""R tests adapter for Otter Assign"""

import re

from dataclasses import dataclass
from jinja2 import Template
from textwrap import indent
from typing import ClassVar

from ..tests_manager import AssignmentTestsManager, TestCase
from ..utils import get_source

from ...test_files.abstract_test import TestFile


# DON'T change this template or the regex that removes hidden tests will break in the R adapter!
OTTR_TEST_FILE_TEMPLATE = Template("""\
test = list(
  name = "{{ name }}",
  cases = list({% for tc in test_cases %}
    ottr::TestCase$new(
      hidden = {% if tc.hidden %}TRUE{% else %}FALSE{% endif %},
      name = {% if tc.name %}"{{ tc.name }}"{% else %}NA{% endif %},
      points = {% if tc.points is none %}NA{% else %}{{ tc.points }}{% endif %},{% if tc.success_message %}
      success_message = "{{ tc.success_message }}",{% endif %}{% if tc.failure_message %}
      failure_message = "{{ tc.failure_message }}",{% endif %}
      code = {
        {{ indent(tc.input, "        ").lstrip() }}{# lstrip so that the first line indent is correct #}
      }
    ){% if not loop.last %},{% endif %}{% endfor %}
  )
)""")
OTTR_TEST_FILE_TEMPLATE.globals['indent'] = indent


@dataclass
class RTestCase(TestCase):
    """
    A dataclass representing a test case for a question in an R assignment.
    """

    name: str
    """the name of the test case"""

    # set the type of output to ClassVar since it's not used for R test cases
    output: ClassVar[None] = None


class RAssignmentTestsManager(AssignmentTestsManager):
    """
    A class for creating and managing test cases for an R assignment.
    """

    def read_test(self, cell, question):
        source = get_source(cell)

        if source[0].lstrip().startswith("#"):
            hidden = bool(re.search(r"\bhidden\b", source[0], flags=re.IGNORECASE))
        else:
            hidden = False

        test_start_line = 0 if hidden else -1

        config, maybe_test_start_line = self._parse_test_config(source)
        test_start_line = maybe_test_start_line if maybe_test_start_line is not None else test_start_line

        test_name = config.get("name", None)
        hidden = config.get("hidden", hidden)
        points = config.get("points", None)
        success_message = config.get("success_message", None)
        failure_message = config.get("failure_message", None)

        self._add_test_case(
            question,
            RTestCase(
                '\n'.join(source[test_start_line+1:]),
                hidden,
                points,
                success_message,
                failure_message,
                test_name,
            ),
        )

    @staticmethod
    def _resolve_test_file_points(total_points, test_cases):
        return TestFile.resolve_test_file_points(total_points, test_cases)

    def _format_test(self, name, points, test_cases):
        template_data = {'name': name, 'test_cases': test_cases}
        return OTTR_TEST_FILE_TEMPLATE.render(**template_data)



import json
import nbformat as nbf
import os
import pytest
import shutil
import tempfile

from otter.check.logs import EventType, Log, LogEntry
from otter.execute import grade_notebook

from ..utils import TestFileManager, write_ok_test


FILE_MANAGER = TestFileManager(__file__)


@pytest.fixture
def temp_dir():
    d = tempfile.mkdtemp()
    yield d
    shutil.rmtree(d)


def test_log_execution(temp_dir):
    """
    Test for ``otter.execute.grade_notebook`` when a log is provided.
    """
    log = Log([
        LogEntry(EventType.INIT),
        LogEntry(EventType.CHECK, question="q1"),
        LogEntry(EventType.CHECK, question="q2"),
        LogEntry(EventType.CHECK, question="q3"),
        LogEntry(EventType.BEGIN_EXPORT),
        LogEntry(EventType.END_EXPORT),
    ])

    log.entries[1].shelve({"a": 1, "b": 2})
    log.entries[2].shelve({"a": 1, "b": 2, "c": None})
    log.entries[3].shelve({"a": 2, "b": 2, "c": 3, "d": 4})

    nb = nbf.v4.new_notebook(cells=[
        nbf.v4.new_code_cell("import pandas as pd"),
        nbf.v4.new_code_cell("from pandas import read_csv"),
        nbf.v4.new_code_cell("raise Exception\na, b = 1, 2"),
        nbf.v4.new_code_cell("raise Exception\nc = None"),
        nbf.v4.new_code_cell("raise Exception\na, c, d = 2, 3, 4"),
    ])

    subm_path = os.path.join(temp_dir, "submission.ipynb")
    nbf.write(nb, subm_path)

    test_dir = os.path.join(temp_dir, "tests")
    os.makedirs(test_dir)

    write_ok_test(os.path.join(test_dir, "q1.py"), ">>> assert a == 1 and b == 2")
    write_ok_test(os.path.join(test_dir, "q2.py"), ">>> assert c is None")
    write_ok_test(os.path.join(test_dir, "q3.py"), ">>> assert a == 2 and c == 3 and d == 4")

    results = grade_notebook(subm_path, test_dir=test_dir, log=log, ignore_errors=False)

    with FILE_MANAGER.open("expected_log_results.json") as f:
        expected_results = json.load(f)

    # fix paths to account for the temp directory
    for d in expected_results.values():
        d["path"] = os.path.join(test_dir, os.path.split(d["path"])[1])

    assert results.to_dict() == expected_results



"""Tests for ``otter.check.logs``"""

import os
import pytest
import sys

from sklearn.linear_model import LinearRegression

from otter.check.logs import Log
from otter.check.notebook import Notebook, _OTTER_LOG_FILENAME
from otter.check.logs import LogEntry, EventType, Log

from ..utils import TestFileManager


FILE_MANAGER = TestFileManager(__file__)


@pytest.fixture(autouse=True)
def cleanup_output(cleanup_enabled):
    yield
    if cleanup_enabled and os.path.isfile(_OTTER_LOG_FILENAME):
        os.remove(_OTTER_LOG_FILENAME)


def test_notebook_check():
    test_directory = FILE_MANAGER.get_path("logs-tests")
    grading_results = {}
    grader = Notebook(tests_dir=test_directory)

    def square(x):
        return x**2

    for test_file in os.listdir(test_directory):
        if os.path.splitext(test_file)[1] != ".py":
            continue

        test_name = os.path.splitext(test_file)[0]
        grading_results[test_name] = grader.check(test_name)

    log = Log.from_file(_OTTER_LOG_FILENAME)

    for question in log.get_questions():
        actual_result = grading_results[question]

        # checking repr since the results __eq__ method is not defined
        assert repr(log.get_results(question)) == repr(actual_result), \
            f"Logged results for {question} are not correct"

        logged_grade = log.get_question_entry(question).get_score_perc()
        assert logged_grade == actual_result.grade, f"Logged results for {question} are not correct"

        # checking repr since the results __eq__ method is not defined
        assert repr(log.get_question_entry(question).get_results()) == repr(actual_result), \
            f"Logged results for {question} are not correct"


def test_shelve():
    """
    tests shelve() and unshelve() which call shelve_environment()
    """

    def square(x):
        return x**2

    import calendar 

    env = {
        "num": 5,
        "func": square,
        "model": LinearRegression(),
        "module": sys,
        "ignored_func": calendar.setfirstweekday
    }

    entry = LogEntry(
        event_type=EventType.CHECK,
        results=[],
        question="foo",
        success=True,
        error=None,
    )


    entry.shelve(env, delete=True, filename=_OTTER_LOG_FILENAME, ignore_modules=['calendar'])
    assert entry.shelf
    assert entry.unshelved == ["module", "ignored_func"]

    entry.flush_to_file(_OTTER_LOG_FILENAME) 

    from math import factorial

    log = Log.from_file(_OTTER_LOG_FILENAME)
    entry = log.get_question_entry("foo")
    env = entry.unshelve()
    assert [*env] == ["num", "func", "model"]

    env_with_factorial = entry.unshelve(dict(factorial = factorial))
    assert "factorial" in env_with_factorial["func"].__globals__
    assert factorial is env_with_factorial["func"].__globals__["factorial"]


def test_log_getitem():
    entry = LogEntry(
        event_type=EventType.AUTH,
        results=[],
        question=None, 
        success=True, 
        error=None,
    )
    entry.flush_to_file(_OTTER_LOG_FILENAME)

    log = Log.from_file(_OTTER_LOG_FILENAME)
    assert log[0].event_type == entry.event_type and log[0].results == entry.results


def test_log_iter():
    entry1 = LogEntry(
        event_type=EventType.CHECK,
        results=[],
        question= "q1", 
        success=True, 
        error=None,
    )

    entry2 = LogEntry(
        event_type=EventType.CHECK,
        results=[],
        question= "q1", 
        success=True, 
        error=None,
    )

    entry3 = LogEntry(
        event_type=EventType.CHECK,
        results=[],
        question= "q2", 
        success=True, 
        error=None,
    )

    entry1.flush_to_file(_OTTER_LOG_FILENAME)
    entry2.flush_to_file(_OTTER_LOG_FILENAME)
    entry3.flush_to_file(_OTTER_LOG_FILENAME)

    log = Log.from_file(_OTTER_LOG_FILENAME)

    log_iter = log.question_iterator()
    assert log_iter.questions == ["q1", "q2"]
    assert next(log_iter).question == entry2.question
    assert next(log_iter).question == entry3.question



"""Tests for ``otter.assign``"""

import nbformat as nbf
import os
import pytest
import shutil

from glob import glob
from unittest import mock

from otter.assign import main as assign
from otter.assign.assignment import Assignment
from otter.assign.question_config import QuestionConfig
from otter.assign.tests_manager import AssignmentTestsManager, TestCase
from otter.generate.token import APIClient
from otter.utils import dump_yaml

from ..utils import assert_dirs_equal, TestFileManager, unzip_to_temp


# prevent pytest from thinking TestCase is a testing class
TestCase.__test__ = False


FILE_MANAGER = TestFileManager(__file__)


@pytest.fixture(autouse=True)
def cleanup_output(cleanup_enabled):
    """
    Removes assign output
    """
    yield
    if cleanup_enabled and os.path.exists(FILE_MANAGER.get_path("output")):
        shutil.rmtree(FILE_MANAGER.get_path("output"))


def check_gradescope_zipfile(path, correct_dir_path):
    """
    Checks that the autograder zip file at ``path`` matches ``correct_dir_path``.
    """
    with unzip_to_temp(path) as unzipped_dir:
        assert_dirs_equal(unzipped_dir, correct_dir_path)


def assign_and_check_output(nb_path, correct_dir, assign_kwargs={}, assert_dirs_equal_kwargs={}):
    """
    Runs Otter Assign and verifies that the output directories are correct.
    """
    output_path = FILE_MANAGER.get_path("output")
    assign(nb_path, output_path, **assign_kwargs)
    assert_dirs_equal(output_path, correct_dir, **assert_dirs_equal_kwargs)


# TODO: refactor existing tests to use this
@pytest.fixture
def generate_master_notebook(cleanup_enabled):
    """
    Yields a function that generates a master notebook file with a specific assignment config and
    returns the path at which the notebook was written. This fixture also deletes the notebook file
    during cleanup.
    """
    nb_path = FILE_MANAGER.get_path("master.ipynb")

    def generate_notebook(assignment_config):
        if os.path.exists(nb_path):
            raise RuntimeError(f"{nb_path} already exists")
        nb = nbf.read(FILE_MANAGER.get_path("master-skeleton.ipynb"), as_version=nbf.NO_CONVERT)
        config_cell = nbf.v4.new_raw_cell("# ASSIGNMENT CONFIG\n" + dump_yaml(assignment_config))
        nb.cells.insert(0, config_cell)

        nbf.write(nb, nb_path)

        return nb_path

    yield generate_notebook

    if cleanup_enabled:
        os.remove(nb_path)


def test_convert_example():
    """
    Checks that otter assign filters and outputs correctly
    """
    assign_and_check_output(
        FILE_MANAGER.get_path("example.ipynb"), 
        FILE_MANAGER.get_path("example-correct"), 
        assign_kwargs=dict(no_run_tests=True),
        assert_dirs_equal_kwargs=dict(variable_path_exts=[".zip"]),
    )

    # check gradescope zip file
    check_gradescope_zipfile(
        glob(FILE_MANAGER.get_path("output/autograder/*.zip"))[0], 
        FILE_MANAGER.get_path("example-autograder-correct"),
    )


def test_exception_example():
    """
    Checks that otter assign filters and outputs correctly
    """
    assign_and_check_output(
        FILE_MANAGER.get_path("exception-example.ipynb"), 
        FILE_MANAGER.get_path("exception-correct"), 
        assign_kwargs=dict(no_run_tests=True),
        assert_dirs_equal_kwargs=dict(variable_path_exts=[".zip"]),
    )


def test_otter_example():
    """
    Checks that otter assign filters and outputs correctly, as well as creates a correct .otter file
    """
    assign_and_check_output(
        FILE_MANAGER.get_path("generate-otter.ipynb"), 
        FILE_MANAGER.get_path("otter-correct"),
        assert_dirs_equal_kwargs=dict(variable_path_exts=[".zip"]),
    )


def test_pdf_example():
    """
    Checks that otter assign filters and outputs correctly, as well as creates a correct .zip file
    along with PDFs
    """
    assign_and_check_output(
        FILE_MANAGER.get_path("generate-pdf.ipynb"),
        FILE_MANAGER.get_path("pdf-correct"),
        assign_kwargs=dict(no_run_tests=True),
        assert_dirs_equal_kwargs=dict(ignore_ext=[".pdf"], variable_path_exts=[".zip"]),
    )


@mock.patch.object(APIClient, "get_token")
def test_gradescope_example(mocked_client):
    """
    Checks that otter assign filters and outputs correctly, as well as creates a correct .zip file
    along with PDFs. Additionally, includes testing Gradescope integration.
    """
    # set a return value that does not match the token in the notebook, so we'll know if APIClient
    # is called
    mocked_client.return_value = 'token'

    assign_and_check_output(
        FILE_MANAGER.get_path("generate-gradescope.ipynb"),
        FILE_MANAGER.get_path("gs-correct"),
        assign_kwargs=dict(no_run_tests=True),
        assert_dirs_equal_kwargs=dict(ignore_ext=[".pdf"], variable_path_exts=[".zip"]),
    )

    # check gradescope zip file
    check_gradescope_zipfile(
        glob(FILE_MANAGER.get_path("output/autograder/*.zip"))[0], 
        FILE_MANAGER.get_path("gs-autograder-correct"),
    )


def test_r_example():
    """
    Checks that otter assign works for R notebooks correctly
    """
    assign_and_check_output(
        FILE_MANAGER.get_path("r-example.ipynb"),
        FILE_MANAGER.get_path("r-correct"),
        assert_dirs_equal_kwargs=dict(ignore_ext=[".pdf"], variable_path_exts=[".zip"]),
    )


def test_rmd_example():
    """
    Checks that otter assign works for Rmd files
    """
    assign_and_check_output(
        FILE_MANAGER.get_path("rmd-example.Rmd"),
        FILE_MANAGER.get_path("rmd-correct"),
        assert_dirs_equal_kwargs=dict(
            ignore_ext=[".pdf"],
            ignore_dirs=["rmd-example-sol_files"],
            variable_path_exts=[".zip"],
        ),
    )

    # check gradescope zip file
    check_gradescope_zipfile(
        glob(FILE_MANAGER.get_path("output/autograder/*.zip"))[0], 
        FILE_MANAGER.get_path("rmd-autograder-correct"),
    )


def test_point_value_rounding():
    """
    Tests that point values are rounded appropriately.
    """
    question = QuestionConfig({"name": "q1", "points": None, "manual": False})
    tests_mgr = AssignmentTestsManager(Assignment())
    for _ in range(11):
        tests_mgr._add_test_case(question, TestCase("", "", False, 4 / 11, "", ""))

    # sum(4 / 11 for _ in range(11)) evaluates to 4.000000000000001 in Python, so this will
    # check that the per-test-case point values are correctly rounded.
    points = tests_mgr.determine_question_point_value(question)
    assert points == 4


def test_determine_question_point_value_error_message():
    """
    Tests that error messages for point value validations contain the question name.
    """
    question = QuestionConfig({"name": "q1", "points": 1, "manual": False})
    tests_mgr = AssignmentTestsManager(Assignment())
    for _ in range(2):
        tests_mgr._add_test_case(question, TestCase("", "", False, 1, "", ""))

    exception = None
    try:
        tests_mgr.determine_question_point_value(question)
    except Exception as e:
        exception = e

    assert str(exception) == "Error in \"q1\" test cases: More points specified in test cases " \
        "than allowed for test"
    assert type(exception) == ValueError


def test_jupyterlite(generate_master_notebook):
    """
    Tests that Otter Assign produces correct notebooks for running on Jupyterlite.
    """
    master_nb_path = generate_master_notebook({
        "runs_on": "jupyterlite",
        "tests": {"url_prefix": "https://domain.tld/tests/"}
    })
    assign_and_check_output(
        master_nb_path,
        FILE_MANAGER.get_path("jupyterlite-correct"),
        assert_dirs_equal_kwargs=dict(variable_path_exts=[".zip"]),
    )


def test_require_no_pdf_ack(generate_master_notebook):
    """
    Tests that Otter Assign produces correct notebooks when configured to require students to
    acknowledge when a PDF cannot be included in their submission zip.
    """
    master_nb_path = generate_master_notebook({
        "export_cell": {
            "require_no_pdf_ack": True,
        },
    })
    assign_and_check_output(
        master_nb_path,
        FILE_MANAGER.get_path("no-pdf-ack-correct"),
        assert_dirs_equal_kwargs=dict(variable_path_exts=[".zip"]),
    )


# TODO: this format of test makes the repo bloated with several golden files. Can these files be
# eliminated?
def test_require_no_pdf_ack_with_message(generate_master_notebook):
    """
    Tests that Otter Assign produces correct notebooks when configured to require students to
    acknowledge when a PDF cannot be included in their submission zip with a custom message.
    """
    master_nb_path = generate_master_notebook({
        "export_cell": {
            "require_no_pdf_ack": {"message": "no pdf could be created"},
        },
    })
    assign_and_check_output(
        master_nb_path,
        FILE_MANAGER.get_path("no-pdf-ack-with-message-correct"),
        assert_dirs_equal_kwargs=dict(variable_path_exts=[".zip"]),
    )



"""Tests for ``otter.grade``"""

import logging
import os
import pandas as pd
import pytest
import re
import shutil
import zipfile

from glob import glob
from python_on_whales import docker
from unittest import mock

from otter.generate import main as generate
from otter.grade import main as grade
from otter.run.run_autograder.autograder_config import AutograderConfig
from otter.utils import loggers

from ..utils import TestFileManager


ASSIGNMENT_NAME = "otter-grade-test"
FILE_MANAGER = TestFileManager(__file__)
AG_ZIP_PATH = FILE_MANAGER.get_path("autograder.zip")
ZIP_SUBM_PATH = "test/subm.zip"


@pytest.fixture(autouse=True)
def cleanup_output(cleanup_enabled):
    yield
    if cleanup_enabled:
        if os.path.exists("test/final_grades.csv"):
            os.remove("test/final_grades.csv")
        if os.path.exists("test/submission_pdfs"):
            shutil.rmtree("test/submission_pdfs")
        if os.path.exists(ZIP_SUBM_PATH):
            os.remove(ZIP_SUBM_PATH)


@pytest.fixture(autouse=True, scope="module")
def generate_zip_file_and_patch_build():
    """
    Generate an autograder zip file for use in these tests and patch
    ``otter.grade.containers.build_image`` with ``build_image_with_local_changes``.
    """
    generate(
        tests_dir = FILE_MANAGER.get_path("tests"), 
        requirements = FILE_MANAGER.get_path("requirements.txt"), 
        output_path = AG_ZIP_PATH,
        config = FILE_MANAGER.get_path("otter_config.json"),
        no_environment = True,
    )

    with loggers.level_context(logging.DEBUG):
        yield

    if os.path.isfile(AG_ZIP_PATH):
        os.remove(AG_ZIP_PATH)


@pytest.fixture
def expected_points():
    test_points = {}
    for test_file in glob(FILE_MANAGER.get_path("tests/*.py")):
        env = {}
        with open(test_file) as f:
            exec(f.read(), env)

        test_points[env['test']['name']] = env['test']['points']

    return test_points


@pytest.mark.slow
@pytest.mark.docker
def test_timeout():
    """
    Check that the notebook ``1min.ipynb`` is killed due to exceeding the defined timeout.
    """
    with pytest.raises(
        Exception, match=r"Executing '[\w./-]*test/test_grade/files/timeout/1min\.ipynb' in docker " \
            "container failed! Exit code: 137"):
        grade(
            name = ASSIGNMENT_NAME,
            paths = [FILE_MANAGER.get_path("timeout/")],
            output_dir = "test/",
            autograder = AG_ZIP_PATH,
            containers = 5,
            timeout = 59,
        )


@pytest.mark.slow
@pytest.mark.docker
def test_network(expected_points):
    """
    Check that the notebook ``network.ipynb`` is unable to do some network requests with disabled
    networking.
    """
    grade(
        name = ASSIGNMENT_NAME,
        paths = [FILE_MANAGER.get_path("network/")],
        output_dir = "test/",
        autograder = AG_ZIP_PATH,
        containers = 5,
        no_network = True,
    )

    df_test = pd.read_csv("test/final_grades.csv")

    # sort by filename
    df_test = df_test.sort_values("file").reset_index(drop=True)

    for _, row in df_test.iterrows():
        for test in expected_points:
            if '/network.ipynb' in row["file"] and ('q2' in test or 'q3' in test):
                assert row[test] == 0, "{} supposed to fail {} but passed".format(row["file"], test)
            else:
                assert row[test] == expected_points[test], "{} supposed to pass {} but failed".format(row["file"], test)


@pytest.mark.slow
@pytest.mark.docker
def test_notebooks_with_pdfs(expected_points):
    """
    Checks that notebooks are graded correctly and that PDFs are generated.
    """
    grade(
        name = ASSIGNMENT_NAME,
        paths = [FILE_MANAGER.get_path("notebooks/")],
        output_dir = "test/",
        autograder = AG_ZIP_PATH,
        containers = 5,
        pdfs = True,
    )

    # read the output and expected output
    df_test = pd.read_csv("test/final_grades.csv")

    # sort by filename
    df_test = df_test.sort_values("file").reset_index(drop=True)
    df_test["failures"] = df_test["file"].apply(lambda x: [int(n) for n in re.split(r"\D+", x) if len(n) > 0])

    # add score sum cols for tests
    for test in expected_points:
        test_cols = [l for l in df_test.columns if bool(re.search(fr"\b{test}\b", l))]
        df_test[test] = df_test[test_cols].sum(axis=1)

    # check point values
    for _, row in df_test.iterrows():
        for test in expected_points:
            if int(re.sub(r"\D", "", test)) in row["failures"]:
                # q6.py has all_or_nothing set to False, so if the hidden tests fail you should get 2.5 points
                if "6H" in row["file"] and "q6" == test:
                    assert row[test] == 2.5, "{} supposed to fail {} but passed".format(row["file"], test)
                else:
                    assert row[test] == 0, "{} supposed to fail {} but passed".format(row["file"], test)
            else:
                assert row[test] == expected_points[test], "{} supposed to pass {} but failed".format(row["file"], test)

    assert os.path.exists("test/submission_pdfs"), "PDF folder is missing"

    # check that an pdf exists for each submission
    dir1_contents, dir2_contents = (
        [os.path.splitext(f)[0] for f in os.listdir(FILE_MANAGER.get_path("notebooks/")) if not (os.path.isdir(os.path.join(FILE_MANAGER.get_path("notebooks/"), f)))],
        [os.path.splitext(f)[0] for f in os.listdir("test/submission_pdfs") if not (os.path.isdir(os.path.join("test/submission_pdfs", f)))],
    )
    assert sorted(dir1_contents) == sorted(dir2_contents), f"'{FILE_MANAGER.get_path('notebooks/')}' and 'test/submission_pdfs' have different contents"


@mock.patch("otter.grade.launch_containers")
def test_single_notebook_grade(mocked_launch_grade):
    """
    Checks that when a single submission is passed to Otter Grade, it returns the percentage score.
    """
    df = pd.DataFrame([{
        "q1": 2.0,
        "q2": 2.0,
        "q3": 2.0,
        "q4": 1.0,
        "q6": 5.0,
        "q2b": 2.0,
        "q7": 1.0,
        "percent_correct": 1.0,
        "file": "passesAll.ipynb",
    }])

    notebook_path = FILE_MANAGER.get_path("notebooks/passesAll.ipynb")

    kw_expected = {
        "num_containers": 1,
        "base_image": "ubuntu:22.04",
        "tag": ASSIGNMENT_NAME,
        "no_kill": False,
        "pdf_dir": None,
        "timeout": None,
        "network": True,
        "config": AutograderConfig(),
    }

    mocked_launch_grade.return_value = [df]

    output = grade(
        name = ASSIGNMENT_NAME,
        paths = [notebook_path],
        output_dir = "test/",
        # the value of the autograder argument doesn't matter, it just needs to be a valid file path
        autograder = notebook_path,
        containers = 1,
    )

    mocked_launch_grade.assert_called_with(notebook_path, [notebook_path], **kw_expected)
    assert output == 1.0


@mock.patch("otter.grade.launch_containers")
def test_config_overrides(mocked_launch_grade):
    """
    Checks that the CLI flags are converted to config overrides correctly.
    """
    mocked_launch_grade.return_value = [pd.DataFrame([{
        "q1": 2.0,
        "q2": 2.0,
        "q3": 2.0,
        "q4": 1.0,
        "q6": 5.0,
        "q2b": 2.0,
        "q7": 1.0,
        "percent_correct": 1.0,
        "file": "passesAll.ipynb",
    }])]

    notebook_path = FILE_MANAGER.get_path("notebooks/passesAll.ipynb")
    grade(
        name = ASSIGNMENT_NAME,
        paths = [notebook_path],
        output_dir = "test/",
        # the value of the autograder argument doesn't matter, it just needs to be a valid file path
        autograder = notebook_path,
        containers = 1,
        pdfs = True,
        ext = "zip",
        debug = True,
    )

    assert mocked_launch_grade.call_args.kwargs["config"].get_user_config() == {
        "zips": True,
        "pdf": True,
        "debug": True,
    }


@pytest.mark.slow
@pytest.mark.docker
def test_config_overrides_integration():
    """
    Checks that overriding otter_config.json configurations with CLI flags works.
    """
    notebook_path = FILE_MANAGER.get_path("notebooks/passesAll.ipynb")
    with zipfile.ZipFile(ZIP_SUBM_PATH, "x") as zf:
        zf.write(notebook_path, arcname="passesAll.ipynb")

    output = grade(
        name = ASSIGNMENT_NAME,
        paths = [ZIP_SUBM_PATH],
        output_dir = "test/",
        # the value of the autograder argument doesn't matter, it just needs to be a valid file path
        autograder = AG_ZIP_PATH,
        ext = "zip",
    )

    assert output == 1.0

    got = pd.read_csv("test/final_grades.csv")
    want = pd.DataFrame([{
        "q1": 0.0,
        "q2": 2.0,
        "q3": 2.0,
        "q4": 1.0,
        "q6": 5.0,
        "q2b": 2.0,
        "q7": 1.0,
        "percent_correct": 1.0,
        "file": ZIP_SUBM_PATH,
    }])

    # Sort the columns by label so the dataframes can be compared with ==.
    got = got.reindex(sorted(got.columns), axis=1)
    want = want.reindex(sorted(want.columns), axis=1)
    
    assert got.equals(want)



"""Local grading of submissions in Docker containers for Otter-Grader"""

import os
import re

from glob import glob
from typing import List, Optional, Tuple, Union

from .containers import launch_containers
from .utils import merge_csv, prune_images

from ..run.run_autograder.autograder_config import AutograderConfig
from ..utils import assert_path_exists, loggers


_ALLOWED_EXTENSIONS = ["ipynb", "py", "Rmd", "R", "r", "zip"]
LOGGER = loggers.get_logger(__name__)


def main(
    *,
    name: Optional[str] = None,
    paths: Optional[Union[List[str], Tuple[str]]] = None,
    output_dir: str = "./",
    autograder: str = "./autograder.zip",
    containers: int = 4, 
    ext: str = "ipynb",
    no_kill: bool = False,
    image: str = "ubuntu:22.04", 
    pdfs: bool = False,
    prune: bool = False,
    force: bool = False,
    timeout: bool = None,
    no_network: bool = False,
    debug: bool = False,
):
    """
    Run Otter Grade.

    Grades a directory of submissions in parallel Docker containers. Results are written as a CSV
    file called ``final_grades.csv`` in ``output_dir``. If ``pdfs`` is true, the PDFs generated
    inside the Docker containers are copied into a subdirectory of ``output_dir`` called
    ``submission_pdfs``.

    If ``prune`` is true, Otter's dangling grading images are pruned and the program exits.

    Args:
        name (``str``): an assignment name to use in the Docker image tag; must be specified unless
            ``prune`` is true
        paths (``list[str] | tuple[str]``): paths to submission files or directories of submissions
            for grading
        output_dir (``str``): path to directory where output should be written
        autograder (``str``): path to an Otter autograder configuration zip file
        containers (``int``): number of containers to run in parallel
        ext (``str``): the submission file extension (to be used in a glob pattern)
        no_kill (``bool``): whether to keep containers after grading is finished
        image (``str``): a Docker image to use as the base image for the grading image
        pdfs (``bool``): whether to copy notebook PDFs out of the containers
        prune (``bool``): whether to prune the grading images; if true, no grading is performed
        force (``bool``): whether to force-prune the images (do not ask for confirmation)
        timeout (``int``): an execution timeout in seconds for each container
        no_network (``bool``): whether to disable networking in the containers
        debug (``bool``): whether to run autograding in debug mode

    Returns:
        ``float | None``: the percentage scored by that submission if a single file was graded
            otherwise ``None``

    Raises:
        ``FileNotFoundError``: if a provided directory or file doesn't exist
        ``ValueError``: if an unsupported extension is passed to ``ext``
    """
    if prune:
        prune_images(force=force)
        return

    if name is None:
        raise ValueError("You must specify an assignment name")
    elif not re.match(r"^[\w\-.]+$", name):
        raise ValueError("Assignment names may only contain letters, nubers, underscores, dashes, and periods")

    if not isinstance(paths, tuple) and not isinstance(paths, list):
        raise TypeError("paths must be a tuple of valid paths")
    elif len(paths) == 0:
        raise ValueError("No paths specified")

    # check file paths
    assert_path_exists([
        (output_dir, True),
        (autograder, False),
    ])

    if ext not in _ALLOWED_EXTENSIONS:
        raise ValueError(f"Invalid submission extension specified: {ext}")

    LOGGER.info("Launching Docker containers")

    pattern = f"*.{ext}"
    submission_paths = []
    for path in paths:
        if os.path.isdir(path):
            submission_paths.extend(glob(os.path.join(path, pattern)))
        else:
            submission_paths.append(path)

    LOGGER.debug(f"Resolved submission paths: {submission_paths}")

    pdf_dir = os.path.join(output_dir, "submission_pdfs") if pdfs else None

    grade_dfs = launch_containers(
        autograder,
        submission_paths,
        num_containers = containers,
        base_image = image,
        tag = name,
        no_kill = no_kill,
        pdf_dir = pdf_dir,
        timeout = timeout,
        network = not no_network,
        config = AutograderConfig({
            "zips": ext == "zip",
            "pdf": pdfs,
            "debug": debug,
        }),
    )

    LOGGER.info("Combining grades and saving")

    # Merge dataframes
    output_df = merge_csv(grade_dfs)
    cols = output_df.columns.tolist()
    output_df = output_df[cols[-1:] + cols[:-1]]

    # write to CSV file
    output_df.to_csv(os.path.join(output_dir, "final_grades.csv"), index=False)

    # return percentage if a single file was graded
    if len(paths) == 1 and os.path.isfile(paths[0]):
        return output_df["percent_correct"][0]



"""Support for notebook metadata test files"""

import json

from .exception_test import ExceptionTestFile
from .ok_test import OKTestFile

from ..utils import NOTEBOOK_METADATA_KEY


class NotebookMetadataExceptionTestFile(ExceptionTestFile):
    """
    A single notebook metadata test file for Otter.

    Tests are defined in the metadata of a jupyter notebook as a JSON object with the ``otter`` key.
    The tests themselves are assumed to be base-64-encoded compiled ``code`` objects from test files.

    .. code-block:: json

    {
        "metadata": {
            "otter": {
                "tests": {
                    "q1": ""
                }
            }
        }
    }

    Args:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        value (``int``, optional): the point value of this test, defaults to 1
        all_or_nothing (``bool``, optional): whether the test should be graded all-or-nothing across
            cases

    Attributes:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        value (``int``): the point value of this test, defaults to 1
        all_or_nothing (``bool``): whether the test should be graded all-or-nothing across
            cases
        passed_all (``bool``): whether all of the test cases were passed
        test_case_results (``list`` of ``TestCaseResult``): a list of results for the test cases in
            ``test_cases``
        grade (``float``): the percentage of ``points`` earned for this test file as a decimal
    """

    @classmethod
    def from_file(cls, path, test_name):
        """
        Parse an exception-based test from a Jupyter notebook's metadata and return an 
        ``ExceptionTestFile``.

        Args:
            path (``str``): the path to the notebook
            test_name (``str``): the name of the test to extract from the metadata

        Returns:
            ``ExceptionTestFile``: the new ``ExceptionTestFile`` object created from the given file
        """
        with open(path, encoding="utf-8") as f:
            nb = json.load(f)

        test_spec = nb["metadata"][NOTEBOOK_METADATA_KEY]["tests"]
        if test_name not in test_spec:
            raise ValueError(f"Test {test_name} not found")

        test_spec = test_spec[test_name]
        return cls.from_string(test_spec, path=path)


class NotebookMetadataOKTestFile(OKTestFile):
    """
    A single notebook metadata test file for Otter.

    Tests are defined in the metadata of a jupyter notebook as a JSON object with the ``otter`` key.
    The tests themselves are OK-formatted.

    .. code-block:: json

    {
        "metadata": {
            "otter": {
                "tests": {
                    "q1": {}
                }
            }
        }
    }

    Args:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        value (``int``, optional): the point value of this test, defaults to 1
        all_or_nothing (``bool``, optional): whether the test should be graded all-or-nothing across
            cases

    Attributes:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        value (``int``): the point value of this test, defaults to 1
        all_or_nothing (``bool``): whether the test should be graded all-or-nothing across
            cases
        passed_all (``bool``): whether all of the test cases were passed
        test_case_results (``list`` of ``TestCaseResult``): a list of results for the test cases in
            ``test_cases``
        grade (``float``): the percentage of ``points`` earned for this test file as a decimal
    """

    @classmethod
    def from_file(cls, path, test_name):
        """
        Parse an OK-formatted test from a Jupyter notebook's metadata and return an ``OKTestFile``.

        Args:
            path (``str``): the path to the notebook
            test_name (``str``): the name of the test to extract from the metadata

        Returns:
            ``OKTestFile``: the new ``OKTestFile`` object created from the given file
        """
        with open(path, encoding="utf-8") as f:
            nb = json.load(f)

        test_spec = nb["metadata"][NOTEBOOK_METADATA_KEY]["tests"]
        if test_name not in test_spec:
            raise ValueError(f"Test {test_name} not found")

        test_spec = test_spec[test_name]

        return cls.from_spec(test_spec, path=path)



"""Utilities for Otter Grade"""

import os
import pandas as pd
import re

from python_on_whales import docker


OTTER_DOCKER_IMAGE_NAME = "otter-grade"


def list_files(path):
    """
    Returns a list of all non-hidden files in a directory

    Args:
        path (``str``): path to a directory

    Returns:
        ``list`` of ``str``: list of filenames in the given directory

    """
    return [file for file in os.listdir(path) if os.path.isfile(os.path.join(path, file)) and file[0] != "."]


def merge_csv(dataframes):
    """
    Merges dataframes along the vertical axis

    Args:
        dataframes (``list`` of ``pandas.core.frame.DataFrame``): list of dataframes with same columns

    Returns:
        ``pandas.core.frame.DataFrame``: A merged dataframe resulting from 'stacking' all input dataframes

    """
    final_dataframe = pd.concat(dataframes, axis=0, join='inner').sort_index()
    return final_dataframe


def prune_images(force=False):
    """
    Prunes all Docker images named ``otter-grade``.
    """
    images = docker.images(OTTER_DOCKER_IMAGE_NAME)
    print("The following images will be deleted:")
    for image in images:
        print(f"    {image.repo_tags[0]}")

    if not force:
        sure = input("Are you sure you want to delete these images? This action cannot be undone [y/N] ")
        sure = bool(re.match(r"ye?s?", sure, flags=re.IGNORECASE))

    else:
        sure = True

    if sure:        
        for image in images:
            image.remove(force=True)

        print("Images deleted.")

    else:
        print("Prune cancelled.")



"""Utilities for Otter Assign"""

import copy
import fica
import json
import os
import pathlib
import re
import shutil

from textwrap import indent

from ..api import grade_submission
from ..generate import main as generate_autograder
from ..run import capture_run_output
from ..utils import (
    get_source,
    loggers,
    NO_PDF_EXPORT_MESSAGE_KEY,
    NOTEBOOK_METADATA_KEY,
    REQUIRE_CONFIRMATION_NO_PDF_EXPORT_KEY,
)


LOGGER = loggers.get_logger(__name__)


class EmptyCellException(Exception):
    """
    Exception for empty cells to indicate deletion
    """


class AssignNotebookFormatException(Exception):
    """
    """
    def __init__(self, message, question, cell_index, *args, **kwargs):
        message += " ("
        if question is not None:
            message += f"question { question.name }, "
        message += f"cell number { cell_index + 1 })"
        super().__init__(message, *args, **kwargs)


def get_notebook_language(nb):
    """
    Parse the notebook kernel language and return it as a string.

    Args:
        nb (``nbformat.NotebookNode``): the notebook

    Returns:
        ``str``: the name of the language as a lowercased string
    """
    return nb["metadata"]["kernelspec"]["language"].lower()


def is_ignore_cell(cell):
    """
    Returns whether the current cell should be ignored

    Args:
        cell (``nbformat.NotebookNode``): a notebook cell

    Returns:
        ``bool``: whether the cell is a ignored
    """
    source = get_source(cell)
    return bool(source and
        re.match(r"(##\s*ignore\s*##\s*|#\s*ignore\s*)", source[0], flags=re.IGNORECASE))


def is_cell_type(cell, cell_type):
    """
    Determine whether a cell is of a specific type.

    Args:
        cell (``nbformat.NotebookNode``): the cell to check
        cell_type (``str``): the cell type to check for

    Returns:
        ``bool``: whether ``cell`` is of type ``cell_type``
    """
    return cell["cell_type"] == cell_type


def remove_output(nb):
    """
    Remove all outputs from a notebook in-place.

    Args:
        nb (``nbformat.NotebookNode``): a notebook
    """
    for cell in nb['cells']:
        if 'outputs' in cell:
            cell['outputs'] = []
        if 'execution_count' in cell:
            cell['execution_count'] = None


def lock(cell):
    """
    Make a cell non-editable and non-deletable in-place.

    Args:
        cell (``nbformat.NotebookNode``): cell to be locked
    """
    m = cell['metadata']
    m["editable"] = False
    m["deletable"] = False


def add_tag(cell, tag):
    """
    Add a tag to a cell, returning a copy.

    Args:
        cell (``nbformat.NotebookNode``): the cell to add a tag to
        tag (``str``): the tag to add

    Returns:
        ``nbformat.NotebookNode``: a copy of ``cell`` with the tag added
    """
    cell = copy.deepcopy(cell)
    if "tags" not in cell["metadata"]:
        cell["metadata"]["tags"] = []
    cell["metadata"]["tags"].append(tag)
    return cell


def has_tag(cell, tag):
    """
    Determine whether a cell has a tag.

    Args:
        cell (``nbformat.NotebookNode``): the cell to check
        tag (``str``): the tag to check for

    Returns:
        ``nbformat.NotebookNode``: whether ``cell`` is tagged with ``tag``
    """
    return tag in cell["metadata"].get("tags", [])


def remove_tag(cell, tag):
    """
    Remove a tag from a cell, returning a copy.

    Args:
        cell (``nbformat.NotebookNode``): the cell to remove a tag from
        tag (``str``): the tag to remove

    Returns:
        ``nbformat.NotebookNode``: a copy of ``cell`` with the tag removed
    """
    cell = copy.deepcopy(cell)
    if "tags" not in cell["metadata"] or tag not in cell["metadata"]["tags"]:
        return cell
    cell["metadata"]["tags"].remove(tag)
    return cell


def str_to_doctest(code_lines, lines):
    """
    Convert a list of lines of Python code ``code_lines`` to the doctest format and appending the
    results to ``lines``.

    Args:
        code_lines (``list[str]``): the code to convert
        lines (``list[str]``): the list to append the converted lines to

    Returns:
        ``list[str]``: a pointer to ``lines``
    """
    if len(code_lines) == 0:
        return lines
    line = code_lines.pop(0)
    if line.startswith(" ") or line.startswith("\t"):
        return str_to_doctest(code_lines, lines + ["... " + line])
    elif bool(re.match(r"^except[\s\w]*:", line)) or line.startswith("elif ") or \
            line.startswith("else:") or line.startswith("finally:"):
        return str_to_doctest(code_lines, lines + ["... " + line])
    elif len(lines) > 0 and lines[-1].strip().endswith("\\"):
        return str_to_doctest(code_lines, lines + ["... " + line])
    else:
        return str_to_doctest(code_lines, lines + [">>> " + line])


def run_tests(assignment, debug=False):
    """
    Grade a notebook and throw an error if it does not receive a perfect score.

    Args:
        assignment (``otter.assgin.assignment.Assignment``): the assignment config
        debug (``bool``): whether to throw errors instead of swallowing them during grading

    Raises:
        ``RuntimeError``: if the grade received by the notebook is not 100%
    """
    with capture_run_output() as run_output:
        results = grade_submission(
            str(assignment.ag_notebook_path),
            str(assignment.ag_zip_path),
            debug=debug,
        )

    LOGGER.debug(f"Otter Run output:\n{run_output.getvalue()}")

    if results.total != results.possible:
        raise RuntimeError(f"Some autograder tests failed in the autograder notebook:\n" + \
            indent(results.summary(), '    '))


def write_otter_config_file(assignment):
    """
    Create an Otter configuration file (a ``.otter`` file) for students to use Otter tools,
    including saving environments and submitting to an Otter Service deployment, using assignment
    configurations. Writes the resulting file to the ``autograder`` and ``student`` subdirectories
    of ``assignment.result``.

    Args:
        assignment (``otter.assign.assignment.Assignment``): the assignment config
    """
    config = {}

    config["notebook"] = assignment.master.name
    config["save_environment"] = assignment.save_environment
    config["ignore_modules"] = assignment.ignore_modules

    if assignment.variables:
        config["variables"] = assignment.variables

    config_name = assignment.master.stem + '.otter'
    with open(assignment.get_ag_path(config_name), "w+") as f:
        json.dump(config, f, indent=4)
    with open(assignment.get_stu_path(config_name), "w+") as f:
        json.dump(config, f, indent=4)


# TODO: update for new assign format
def run_generate_autograder(assignment, gs_username, gs_password, plugin_collection=None):
    """
    Run Otter Generate on the autograder directory to generate a Gradescope zip file. Relies on
    configurations in ``assignment.generate``.

    Args:
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        gs_username (``str``): Gradescope username for token generation
        gs_password (``str``): Gradescope password for token generation
        plugin_collection (``otter.plugins.PluginCollection``, optional): a plugin collection to pass
            to Otter Generate
    """
    curr_dir = os.getcwd()
    os.chdir(str(assignment.get_ag_path()))

    # use temp tests dir
    test_dir = "tests"
    if assignment.is_python and not assignment.tests.files and assignment._temp_test_dir is None:
        raise RuntimeError("Failed to create temp tests directory for Otter Generate")

    elif assignment.is_python and not assignment.tests.files:
        test_dir = str(assignment._temp_test_dir)

    requirements = None
    if assignment.requirements:
        if assignment.is_r:
            requirements = 'requirements.R'
        else:
            requirements = 'requirements.txt'

    files = []
    if assignment.files:
        files += assignment.files

    if assignment.autograder_files:
        ag_dir = os.getcwd()
        os.chdir(curr_dir)
        output_dir = assignment.get_ag_path()

        # copy files
        for file in assignment.autograder_files:

            # if a directory, copy the entire dir
            if os.path.isdir(file):
                shutil.copytree(file, str(output_dir / os.path.basename(file)))

            else:
                # check that file is in subdir
                assert os.getcwd() in os.path.abspath(file), \
                    f"{file} is not in a subdirectory of the master notebook directory"
                file_path = pathlib.Path(file).resolve()
                rel_path = file_path.parent.relative_to(pathlib.Path(os.getcwd()))
                os.makedirs(output_dir / rel_path, exist_ok=True)
                shutil.copy(file, str(output_dir / rel_path))

        os.chdir(ag_dir)

        files += assignment.autograder_files

    otter_config = assignment.get_otter_config()
    if otter_config:
        # TODO: move this filename into a global variable somewhere and remove all of the places
        # it's hardcoded
        with open("otter_config.json", "w+") as f:
            json.dump(otter_config, f, indent=2)

    # TODO: change generate_autograder so that only necessary kwargs are needed
    generate_autograder(
        tests_dir=test_dir,
        output_path=assignment.ag_zip_name,
        config="otter_config.json" if otter_config else None,
        lang="python" if assignment.is_python else "r",
        requirements=requirements,
        overwrite_requirements=assignment.overwrite_requirements,
        environment="environment.yml" if assignment.environment else None,
        no_environment=False,
        username=gs_username,
        password=gs_password,
        files=files,
        plugin_collection=plugin_collection,
        assignment=assignment,
        python_version=assignment.get_python_version(),
        channel_priority_strict=assignment.channel_priority_strict,
    )

    # clean up temp tests dir
    if assignment._temp_test_dir is not None:
        shutil.rmtree(str(assignment._temp_test_dir))

    os.chdir(curr_dir)


def add_assignment_name_to_notebook(nb, assignment):
    """
    Add the assignment name from the assignment config to the provided notebook's metadata in-place.

    If ``assignment`` has a name, the name is added to the notebook metadata, as
    ``nb["metadata"]["otter"]["assignment_name"]``.

    Args:
        nb (``nbformat.NotebookNode``): the notebook to add the name to
        assignment (``otter.assign.assignment.Assignment``): the assignment config
    """
    if assignment.name is not None:
        if NOTEBOOK_METADATA_KEY not in nb["metadata"]:
            nb["metadata"][NOTEBOOK_METADATA_KEY] = {}
        nb["metadata"][NOTEBOOK_METADATA_KEY]["assignment_name"] = assignment.name


def add_require_no_pdf_ack_to_notebook(nb, assignment):
    """
    Add the no PDF ACK configurtion from the assignment config to the provided notebook's metadata
    in-place.

    Args:
        nb (``nbformat.NotebookNode``): the notebook to add the name to
        assignment (``otter.assign.assignment.Assignment``): the assignment config
    """
    if assignment.export_cell and assignment.export_cell.require_no_pdf_ack:
        if NOTEBOOK_METADATA_KEY not in nb["metadata"]:
            nb["metadata"][NOTEBOOK_METADATA_KEY] = {}
        nb["metadata"][NOTEBOOK_METADATA_KEY][REQUIRE_CONFIRMATION_NO_PDF_EXPORT_KEY] = True
        if isinstance(assignment.export_cell.require_no_pdf_ack, fica.Config):
            nb["metadata"][NOTEBOOK_METADATA_KEY][NO_PDF_EXPORT_MESSAGE_KEY] = \
                assignment.export_cell.require_no_pdf_ack.message


def remove_cell_ids_if_applicable(nb):
    """
    Remove all cell IDs from a notebook in-place iff the nbformat of the notebook is < 4.5.

    Args:
        nb (``nbformat.NotebookNode``): a notebook
    """
    if nb["nbformat"] < 4 or (nb["nbformat"] == 4 and nb["nbformat_minor"] < 5):
        for cell in nb.cells:
            if "id" in cell:
                cell.pop("id")



"""Tests for ``otter.api``"""

import os

from unittest import mock

from otter.api import grade_submission


@mock.patch("otter.api.run_grader")
@mock.patch("otter.api.tempfile")
@mock.patch("otter.api.redirect_stdout")
def test_grade_submission(mocked_redirect, mocked_tempfile, mocked_run, tmp_path):
    """
    Tests for ``otter.api.grade_submission``.
    """
    subm_path = "foo.ipynb"
    temp_dir = str(tmp_path / "autograder")
    os.mkdir(temp_dir)

    mocked_tempfile.mkdtemp.return_value = temp_dir

    grade_submission(subm_path)

    mocked_run.assert_called_with(
        subm_path,
        autograder="autograder.zip",
        output_dir=temp_dir,
        no_logo=True,
        debug=False,
    )
    mocked_redirect.assert_not_called()

    os.mkdir(temp_dir)
    grade_submission(subm_path, quiet=True)

    mocked_redirect.assert_called()



"""Abstract runner class for the autograder"""

import json
import os
import shutil
import tempfile

from abc import ABC, abstractmethod

from ..autograder_config import AutograderConfig
from ..utils import OtterRuntimeError, print_output, write_blank_page_to_stare_at_before_you

from ....utils import NOTEBOOK_METADATA_KEY


class AbstractLanguageRunner(ABC):
    """
    A class defining the logic of running the autograder and generating grading results.

    Args:
        otter_config (``dict[str, object]``): user-specified configurations to override the defaults
        **kwargs: other user-specified configurations to override the defaults and values specified
            in ``otter_config``
    """

    ag_config: AutograderConfig
    """the autograder config"""

    def __init__(self, otter_config, **kwargs):
        self.ag_config = AutograderConfig({**otter_config, **kwargs})

    @staticmethod
    def determine_language(otter_config, **kwargs):
        """
        Determine the language of the assignment based on user-specified configurations.
        """
        # TODO: use fica.Key.get_default when available
        return kwargs.get("lang", otter_config.get("lang", AutograderConfig.lang.get_value()))

    def get_option(self, option):
        """
        Return the value of a configuration, including defaults.
        """
        return self.ag_config[option]

    def get_config(self):
        """
        Return the autograder config.
        """
        return self.ag_config

    def prepare_files(self):
        """
        Copies tests and support files needed for running the autograder.

        When this method is invoked, the working directory is assumed to already be 
        ``self.ag_config.autograder_dir``.
        """
        # put files into submission directory
        if os.path.exists("./source/files"):
            for file in os.listdir("./source/files"):
                fp = os.path.join("./source/files", file)
                if os.path.isdir(fp):
                    if not os.path.exists(os.path.join("./submission", os.path.basename(fp))):
                        shutil.copytree(fp, os.path.join("./submission", os.path.basename(fp)))
                else:
                    shutil.copy(fp, "./submission")

        # copy the tests directory
        if os.path.exists("./submission/tests"):
            shutil.rmtree("./submission/tests")
        shutil.copytree("./source/tests", "./submission/tests")

    def validate_assignment_name(self, got):
        """
        Raise an ``OtterRuntimeError`` if the assignment name of the notebook is invalid.

        If no assignment name was provided in the configurations, no action is taken.

        Args:
            got (``str | None``): the assignment name of the submission or ``None`` if it didn't
                have one

        Raises:
            ``otter.run.run_autograder.utils.OtterRuntimeError``: if the name is invalid and
                grading should be aborted
        """
        if self.ag_config.assignment_name and got != self.ag_config.assignment_name:
            message = f"Received submission for assignment '{got}' (this is assignment " \
                f"'{self.ag_config.assignment_name}')"
            raise OtterRuntimeError(message)

    def get_notebook_assignment_name(self, nb):
        """
        Get the assignment name in the metadata of the provided notebook, if any.

        Args:
            nb (``nbformat.NotebookNode``): the notebook

        Returns:
            ``str | None``: the assignment name of the notebook, if any
        """
        if NOTEBOOK_METADATA_KEY not in nb["metadata"]:
            return None

        return nb["metadata"][NOTEBOOK_METADATA_KEY].get("assignment_name", None)

    def write_and_maybe_submit_pdf(self, client, submission_path, submit, scores):
        """
        Upload a PDF to a Gradescope assignment for manual grading.

        This method calls ``self.write_pdf`` to generate the PDF before submitting it, and catches
        any errors thrown in that function.

        This method should be called, if appropriate, by ``self.run``.

        Args:
            client (``otter.generate.token.APIClient``): the Gradescope client
            submission_path (``str``): path to the submission
            submit (``bool``): whether to submit the PDF with ``client``
            scores (``otter.test_files.GradingResults``): the grading results (so an error can be
                appended if needed)
        """
        try:
            pdf_path = self.write_pdf(submission_path)

            if submit:
                self.submit_pdf(client, pdf_path)

        except Exception as e:
            print_output(f"\n\nError encountered while generating and submitting PDF:\n{e}")
            scores.set_pdf_error(e)

            if self.ag_config.submit_blank_pdf_on_export_failure and submit:
                print_output("\nUploading a blank PDF due to export failure")
                with tempfile.NamedTemporaryFile(suffix=".pdf") as ntf:
                    write_blank_page_to_stare_at_before_you(ntf.name)
                    self.submit_pdf(client, ntf.name)

    def submit_pdf(self, client, pdf_path):
        """
        Upload the PDF at ``pdf_path`` to the Gradescope assignment specified in the config. Does
        not check whether the assignment configuration is valid.

        Args:
            client (``otter.generate.token.APIClient``): the Gradescope client
            pdf_path (``str``): path to the PDF file to upload
        """
        # get student email
        with open("../submission_metadata.json", encoding="utf-8") as f:
            metadata = json.load(f)

        student_emails = []
        for user in metadata.get("users", []):
            student_emails.append(user["email"])

        for student_email in student_emails:
            client.upload_pdf_submission(
                self.ag_config.course_id,
                self.ag_config.assignment_id,
                student_email,
                pdf_path,
            )

        print_output("\n\nSuccessfully uploaded submissions for: {}".format(
            ", ".join(student_emails)))

    @abstractmethod
    def validate_submission(self, submission_path):
        """
        Validate the submission, raising an error/warning if appropriate.

        This method should be invoked as part of the implementation of another abstract method
        (either ``resolve_submission_path`` or ``run``).

        Args:
            submission_path (``str``): the path to the submission file
        """
        ...

    @abstractmethod
    def resolve_submission_path(self):
        """
        Determine the path to the submission file, performing any necessary transformations on the
        file.

        When this method is invoked, the working directory is assumed to already be 
        ``{self.ag_config.autograder_dir}/submission``.
        """
        ...

    @abstractmethod
    def write_pdf(self, submission_path):
        """
        Convert the submission to a PDF, returning the path to the PDF file.

        Args:
            submission_path (``str``): the path to the submission

        Returns:
            ``str``: the path to the generated PDF
        """
        ...

    @abstractmethod
    def run(self):
        """
        Run the autograder according to the configurations in ``self.ag_config``.

        When this method is invoked, the working directory is assumed to already be 
        ``self.ag_config.autograder_dir``.

        Returns:
            ``otter.test_files.GradingResults``: the results from grading the submission
        """
        ...



import pytest

from unittest.mock import patch, MagicMock

from otter.execute.checker import Checker
from otter.test_files import OKTestFile


@pytest.fixture
def mocked_create_test_file():
    with patch("otter.execute.checker.create_test_file") as m:
        yield m


@pytest.fixture(autouse=True)
def reset_checker():
    Checker.clear_results()
    Checker.disable_tracking()


class TestChecker:
    """
    Tests for ``otter.execute.checker.Checker``.
    """

    def test_tracking_disabled(self, mocked_create_test_file):
        res = Checker.get_results()
        Checker.check("")
        assert len(res) == 0

    def test_tracking_enabled(self, mocked_create_test_file):
        res = Checker.get_results()
        Checker.enable_tracking()
        Checker.check("")
        assert len(res) == 1

    def test_tracking_toggles(self, mocked_create_test_file):
        res = Checker.get_results()
        Checker.check("")
        assert len(res) == 0

        Checker.enable_tracking()
        Checker.check("")
        assert len(res) == 1

        Checker.check("")
        assert len(res) == 2

        Checker.disable_tracking()
        Checker.check("")
        assert len(res) == 2

    def test_get_results_and_clear_results(self, mocked_create_test_file):
        res = Checker.get_results()
        assert res is Checker.get_results()

        Checker.clear_results()
        assert res is not Checker.get_results()

    @patch("otter.execute.checker.inspect.currentframe")
    def test_check_path_only(self, mocked_currentframe, mocked_create_test_file):
        path = "foo.ipynb"

        ret = Checker.check(path)

        mocked_create_test_file.assert_called_once_with(path, test_name=None)
        mocked_currentframe.assert_called_once()
        mocked_create_test_file.return_value.run.assert_called_once_with(
            mocked_currentframe.return_value.f_back.f_globals)
        assert ret is mocked_create_test_file.return_value

    @patch("otter.execute.checker.inspect.currentframe")
    def test_check_path_and_test_name(self, mocked_currentframe, mocked_create_test_file):
        path, test_name = "foo.ipynb", "q1"

        ret = Checker.check(path, test_name)

        mocked_create_test_file.assert_called_once_with(path, test_name=test_name)
        mocked_currentframe.assert_called_once()
        mocked_create_test_file.return_value.run.assert_called_once_with(
            mocked_currentframe.return_value.f_back.f_globals)
        assert ret is mocked_create_test_file.return_value

    @patch("otter.execute.checker.inspect.currentframe")
    def test_check_path_and_global_env(self, mocked_currentframe, mocked_create_test_file):
        path, global_env = "foo.ipynb", {}

        ret = Checker.check(path, global_env=global_env)

        mocked_create_test_file.assert_called_once_with(path, test_name=None)
        mocked_currentframe.assert_not_called()
        mocked_create_test_file.return_value.run.assert_called_once_with(global_env)
        assert ret is mocked_create_test_file.return_value

    @patch("otter.execute.checker.inspect.currentframe")
    @patch.object(Checker, "check")
    def test_check_if_not_already_checked(self, mocked_check, mocked_currentframe):
        Checker._test_files = [
            OKTestFile("", "q1.py", []),
            OKTestFile("", "tests/q2.py", []),
            OKTestFile("", "/path/to/tests/q3.py", []),
        ]

        for t in [
            {"path": "q1.py", "want_call": False},
            {"path": "q2.py", "want_call": False},
            {"path": "q3.py", "want_call": False},
            {"path": "/path/to/tests/q2.py", "want_call": False},
            {"path": "tests/q3.py", "want_call": False},
            {"path": "q4.py", "want_call": True},
            {"path": "tests/q4.py", "want_call": True},
            {"path": "/path/to/tests/q4.py", "want_call": True},
        ]:
            mocked_check.reset_mock()

            ret = Checker.check_if_not_already_checked(t["path"], t.get("global_env"))

            if t["want_call"]:
                global_env = t.get("global_env", mocked_currentframe.return_value.f_back.f_globals)
                mocked_check.assert_called_once_with(t["path"], global_env=global_env)
                assert ret is mocked_check.return_value
            else:
                mocked_check.assert_not_called()



"""Tests for ``otter.cli``"""

import logging
import os
import pytest

from click.testing import CliRunner
from unittest import mock

from otter import __version__
from otter.cli import cli
from otter.generate import main as generate
from otter.grade import _ALLOWED_EXTENSIONS, main as grade
from otter.run import main as run


def assert_cli_result(result, expect_error):
    """
    Asserts that the ``CliRunner`` result exited with code 0.
    """
    assert result.exit_code != 0 if expect_error else result.exit_code == 0, \
        result.stdout_bytes.decode("utf-8")


@pytest.fixture()
def run_cli():
    """
    Create a function to invoke Otter's CLI with the specified command.

    Yields the function in the ``click.testing.CliRunner``'s ``isolated_filesystem`` context.

    Yields:
        ``callable[[list[str]], click.testing.Result]``: the function to invoke the CLI
    """
    runner = CliRunner()
    with runner.isolated_filesystem():
        yield lambda cmd: runner.invoke(cli, cmd)


@mock.patch("otter.cli.print_version_info")
def test_version(mocked_version, run_cli):
    """
    Tests the ``otter --version`` CLI command.
    """
    result = run_cli([])
    assert_cli_result(result, expect_error=False)
    mocked_version.assert_not_called()

    result = run_cli(["--version"])
    assert_cli_result(result, expect_error=False)
    mocked_version.assert_called_once_with(logo=True)


@mock.patch("otter.cli.assign")
@mock.patch("otter.cli.loggers")
def test_verbosity(mocked_loggers, _, run_cli):
    """
    Tests setting the verbosity of Otter's message logging system.
    """
    open("foo.ipynb", "w+").close()

    run_cli(["assign", "foo.ipynb", "dist"])
    mocked_loggers.set_level.assert_called_with(logging.WARNING)

    run_cli(["assign", "foo.ipynb", "dist", "-v"])
    mocked_loggers.set_level.assert_called_with(logging.INFO)

    run_cli(["assign", "foo.ipynb", "dist", "-vv"])
    mocked_loggers.set_level.assert_called_with(logging.DEBUG)

    run_cli(["assign", "foo.ipynb", "dist", "-vvv"])
    mocked_loggers.set_level.assert_called_with(logging.DEBUG)


@mock.patch("otter.cli.assign")
def test_assign(mocked_assign, run_cli):
    """
    Tests the ``otter assign`` CLI command.
    """
    master, result = "foo.ipynb", "dist"
    cmd_start = ["assign", master, result]

    open(master, "w+").close()

    std_kwargs = dict(
        master=master,
        result=result,
        no_run_tests=False,
        no_pdfs=False,
        username=None,
        password=None,
        debug=False,
    )

    result = run_cli([*cmd_start])
    assert_cli_result(result, expect_error=False)
    mocked_assign.assert_called_with(**std_kwargs)

    result = run_cli([*cmd_start, "--no-run-tests"])
    assert_cli_result(result, expect_error=False)
    mocked_assign.assert_called_with(**{**std_kwargs, "no_run_tests": True})

    result = run_cli([*cmd_start, "--no-pdfs"])
    assert_cli_result(result, expect_error=False)
    mocked_assign.assert_called_with(**{**std_kwargs, "no_pdfs": True})

    un, pw = "foo", "bar"
    result = run_cli([*cmd_start, "--username", un, "--password", pw])
    assert_cli_result(result, expect_error=False)
    mocked_assign.assert_called_with(**{**std_kwargs, "username": un, "password": pw})

    result = run_cli([*cmd_start, "--debug"])
    assert_cli_result(result, expect_error=False)
    mocked_assign.assert_called_with(**{**std_kwargs, "debug": True})

    # test invalid calls
    mocked_assign.reset_mock()

    result = run_cli(["assign", "bar.ipynb", result])
    assert_cli_result(result, expect_error=True)
    mocked_assign.assert_not_called()

    os.mkdir("bar")
    result = run_cli(["assign", "bar", result])
    assert_cli_result(result, expect_error=True)
    mocked_assign.assert_not_called()

    result = run_cli(["assign", "foo.ipynb"])
    assert_cli_result(result, expect_error=True)
    mocked_assign.assert_not_called()


@mock.patch("otter.cli.check")
def test_check(mocked_check, run_cli):
    """
    Tests the ``otter check`` CLI command.
    """
    file = "foo.py"
    cmd_start = ["check", file]

    open(file, "w+").close()
    os.mkdir("tests")

    std_kwargs = dict(
        file=file,
        tests_path="./tests",
        question=None,
        seed=None,
    )

    result = run_cli([*cmd_start])
    assert_cli_result(result, expect_error=False)
    mocked_check.assert_called_with(**std_kwargs)

    result = run_cli([*cmd_start, "-q", "q1"])
    assert_cli_result(result, expect_error=False)
    mocked_check.assert_called_with(**{**std_kwargs, "question": "q1"})

    result = run_cli([*cmd_start, "--question", "q1"])
    assert_cli_result(result, expect_error=False)
    mocked_check.assert_called_with(**{**std_kwargs, "question": "q1"})

    os.mkdir("tests2")
    result = run_cli([*cmd_start, "-t", "tests2"])
    assert_cli_result(result, expect_error=False)
    mocked_check.assert_called_with(**{**std_kwargs, "tests_path": "tests2"})

    result = run_cli([*cmd_start, "--tests-path", "tests2"])
    assert_cli_result(result, expect_error=False)
    mocked_check.assert_called_with(**{**std_kwargs, "tests_path": "tests2"})

    result = run_cli([*cmd_start, "--seed", "1"])
    assert_cli_result(result, expect_error=False)
    mocked_check.assert_called_with(**{**std_kwargs, "seed": 1})

    # test invalid calls
    mocked_check.reset_mock()

    result = run_cli(["check"])
    assert_cli_result(result, expect_error=True)
    mocked_check.assert_not_called()

    result = run_cli(["check", "tests2"])
    assert_cli_result(result, expect_error=True)
    mocked_check.assert_not_called()

    result = run_cli(["check", file, "-t", "tests3"])
    assert_cli_result(result, expect_error=True)
    mocked_check.assert_not_called()

    open("foo.txt", "w+").close()
    result = run_cli(["check", file, "-t", "foo.txt"])
    assert_cli_result(result, expect_error=True)
    mocked_check.assert_not_called()

    result = run_cli(["check", file, "--seed", "foo"])
    assert_cli_result(result, expect_error=True)
    mocked_check.assert_not_called()


@mock.patch("otter.cli.export")
def test_export(mocked_export, run_cli):
    """
    Tests the ``otter export`` CLI command.
    """
    src = "foo.ipynb"
    cmd_start = ["export", src]

    open(src, "w+").close()

    std_kwargs = dict(
        src=src,
        dest=None,
        exporter=None,
        filtering=False,
        pagebreaks=False,
        save=False,
        xecjk=False,
    )

    result = run_cli([*cmd_start])
    assert_cli_result(result, expect_error=False)
    mocked_export.assert_called_with(**std_kwargs)

    result = run_cli([*cmd_start, "foo.pdf"])
    assert_cli_result(result, expect_error=False)
    mocked_export.assert_called_with(**{**std_kwargs, "dest": "foo.pdf"})

    result = run_cli([*cmd_start, "--filtering"])
    assert_cli_result(result, expect_error=False)
    mocked_export.assert_called_with(**{**std_kwargs, "filtering": True})

    result = run_cli([*cmd_start, "--pagebreaks"])
    assert_cli_result(result, expect_error=False)
    mocked_export.assert_called_with(**{**std_kwargs, "pagebreaks": True})

    result = run_cli([*cmd_start, "-s"])
    assert_cli_result(result, expect_error=False)
    mocked_export.assert_called_with(**{**std_kwargs, "save": True})

    result = run_cli([*cmd_start, "--save"])
    assert_cli_result(result, expect_error=False)
    mocked_export.assert_called_with(**{**std_kwargs, "save": True})

    result = run_cli([*cmd_start, "--xecjk"])
    assert_cli_result(result, expect_error=False)
    mocked_export.assert_called_with(**{**std_kwargs, "xecjk": True})

    result = run_cli([*cmd_start, "-e", "latex"])
    assert_cli_result(result, expect_error=False)
    mocked_export.assert_called_with(**{**std_kwargs, "exporter": "latex"})

    result = run_cli([*cmd_start, "--exporter", "html"])
    assert_cli_result(result, expect_error=False)
    mocked_export.assert_called_with(**{**std_kwargs, "exporter": "html"})

    # test invalid calls
    mocked_export.reset_mock()

    result = run_cli(["export"])
    assert_cli_result(result, expect_error=True)
    mocked_export.assert_not_called()

    result = run_cli(["export", "bar.ipynb"])
    assert_cli_result(result, expect_error=True)
    mocked_export.assert_not_called()

    os.mkdir("bar")
    result = run_cli(["export", "bar"])
    assert_cli_result(result, expect_error=True)
    mocked_export.assert_not_called()

    result = run_cli(["export", src, "-e", "foo"])
    assert_cli_result(result, expect_error=True)
    mocked_export.assert_not_called()


@mock.patch("otter.cli.generate")
def test_generate(mocked_generate, run_cli):
    """
    Tests the ``otter generate`` CLI command.
    """
    cmd_start = ["generate"]

    os.mkdir("tests")
    open("otter_config.json", "w+").close()
    open("requirements.txt", "w+").close()
    open("environment.yml", "w+").close()

    std_kwargs = dict(
        **generate.__kwdefaults__,
    )
    std_kwargs["files"] = tuple()
    std_kwargs["channel_priority_strict"] = False
    std_kwargs.pop("assignment")
    std_kwargs.pop("plugin_collection")

    result = run_cli([*cmd_start])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**std_kwargs)

    os.mkdir("tests2")
    result = run_cli([*cmd_start, "-t", "tests2"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "tests_dir": "tests2"})

    result = run_cli([*cmd_start, "--tests-dir", "tests2"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "tests_dir": "tests2"})

    result = run_cli([*cmd_start, "-o", "output.zip"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "output_path": "output.zip"})

    result = run_cli([*cmd_start, "--output-path", "output.zip"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "output_path": "output.zip"})

    result = run_cli([*cmd_start, "-c", "otter_config.json"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "config": "otter_config.json"})

    result = run_cli([*cmd_start, "--config", "otter_config.json"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "config": "otter_config.json"})

    result = run_cli([*cmd_start, "--no-config"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "no_config": True})

    result = run_cli([*cmd_start, "-r", "requirements.txt"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "requirements": "requirements.txt"})

    result = run_cli([*cmd_start, "--requirements", "requirements.txt"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "requirements": "requirements.txt"})

    result = run_cli([*cmd_start, "--no-requirements"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "no_requirements": True})

    result = run_cli([*cmd_start, "-e", "environment.yml"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "environment": "environment.yml"})

    result = run_cli([*cmd_start, "--environment", "environment.yml"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "environment": "environment.yml"})

    result = run_cli([*cmd_start, "--no-environment"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "no_environment": True})

    result = run_cli([*cmd_start, "-l", "python"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "lang": "python"})

    result = run_cli([*cmd_start, "--lang", "r"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "lang": "r"})

    result = run_cli([*cmd_start, "--username", "foo", "--password", "bar"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "username": "foo", "password": "bar"})

    result = run_cli([*cmd_start, "--token", "abc123"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "token": "abc123"})

    result = run_cli([*cmd_start, "foo", "bar", "baz"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "files": ("foo", "bar", "baz")})

    result = run_cli([*cmd_start, "--channel-priority-strict"])
    assert_cli_result(result, expect_error=False)
    mocked_generate.assert_called_with(**{**std_kwargs, "channel_priority_strict": True})

    # test invalid calls
    mocked_generate.reset_mock()

    result = run_cli(["generate", "-t", "tests3"])
    assert_cli_result(result, expect_error=True)
    mocked_generate.assert_not_called()

    result = run_cli(["generate", "-t", "otter_config.json"])
    assert_cli_result(result, expect_error=True)
    mocked_generate.assert_not_called()

    result = run_cli(["generate", "-c", "tests"])
    assert_cli_result(result, expect_error=True)
    mocked_generate.assert_not_called()

    result = run_cli(["generate", "-c", "bar.txt"])
    assert_cli_result(result, expect_error=True)
    mocked_generate.assert_not_called()

    result = run_cli(["generate", "-r", "tests"])
    assert_cli_result(result, expect_error=True)
    mocked_generate.assert_not_called()

    result = run_cli(["generate", "-r", "bar.txt"])
    assert_cli_result(result, expect_error=True)
    mocked_generate.assert_not_called()

    result = run_cli(["generate", "-e", "tests"])
    assert_cli_result(result, expect_error=True)
    mocked_generate.assert_not_called()

    result = run_cli(["generate", "-e", "bar.txt"])
    assert_cli_result(result, expect_error=True)
    mocked_generate.assert_not_called()


@mock.patch("otter.cli.grade")
def test_grade(mocked_grade, run_cli):
    """
    Tests the ``otter grade`` CLI command.
    """
    cmd_start = ["grade"]

    open("autograder.zip", "wb+").close()

    std_kwargs = dict(
        **grade.__kwdefaults__,
    )
    std_kwargs["paths"] = ()

    result = run_cli([*cmd_start])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**std_kwargs)

    os.mkdir("notebooks")
    result = run_cli([*cmd_start, "notebooks"])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**{**std_kwargs, "paths": ("notebooks", )})

    open("foo.ipynb", "w+").close()
    result = run_cli([*cmd_start, "foo.ipynb"])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**{**std_kwargs, "paths": ("foo.ipynb",)})

    open("bar.ipynb", "w+").close()
    open("baz.ipynb", "w+").close()
    result = run_cli([*cmd_start, "foo.ipynb", "bar.ipynb", "baz.ipynb"])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**{**std_kwargs, "paths": ("foo.ipynb", "bar.ipynb", "baz.ipynb")})

    result = run_cli([*cmd_start, "-n", "hw01"])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**{**std_kwargs, "name": "hw01"})

    result = run_cli([*cmd_start, "--name", "hw01"])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**{**std_kwargs, "name": "hw01"})

    open("ag2.zip", "wb+").close()
    result = run_cli([*cmd_start, "-a", "ag2.zip"])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**{**std_kwargs, "autograder": "ag2.zip"})

    result = run_cli([*cmd_start, "--autograder", "ag2.zip"])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**{**std_kwargs, "autograder": "ag2.zip"})

    os.mkdir("output")
    result = run_cli([*cmd_start, "-o", "output"])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**{**std_kwargs, "output_dir": "output"})

    result = run_cli([*cmd_start, "--output-dir", "output"])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**{**std_kwargs, "output_dir": "output"})

    for ext in _ALLOWED_EXTENSIONS:
        result = run_cli([*cmd_start, "--ext", ext])
        assert_cli_result(result, expect_error=False)
        mocked_grade.assert_called_with(**{**std_kwargs, "ext": ext})

    result = run_cli([*cmd_start, "--pdfs"])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**{**std_kwargs, "pdfs": True})

    result = run_cli([*cmd_start, "--containers", "10"])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**{**std_kwargs, "containers": 10})

    result = run_cli([*cmd_start, "--image", "foo"])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**{**std_kwargs, "image": "foo"})

    result = run_cli([*cmd_start, "--timeout", "300"])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**{**std_kwargs, "timeout": 300})

    result = run_cli([*cmd_start, "--no-network"])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**{**std_kwargs, "no_network": True})

    result = run_cli([*cmd_start, "--no-kill"])
    assert_cli_result(result, expect_error=False)
    mocked_grade.assert_called_with(**{**std_kwargs, "no_kill": True})

    # test invalid calls
    mocked_grade.reset_mock()

    result = run_cli([*cmd_start, "quux.ipynb"])
    assert_cli_result(result, expect_error=True)
    mocked_grade.assert_not_called()

    result = run_cli([*cmd_start, "--ext", "foo"])
    assert_cli_result(result, expect_error=True)
    mocked_grade.assert_not_called()

    result = run_cli([*cmd_start, "--containers", "foo"])
    assert_cli_result(result, expect_error=True)
    mocked_grade.assert_not_called()

    result = run_cli([*cmd_start, "--timeout", "foo"])
    assert_cli_result(result, expect_error=True)
    mocked_grade.assert_not_called()


@mock.patch("otter.cli.run")
def test_run(mocked_run, run_cli):
    """
    Tests the ``otter run`` CLI command.
    """
    cmd_start = ["run", "foo.ipynb"]

    open("foo.ipynb", "w+").close()
    open("autograder.zip", "wb+").close()

    std_kwargs = dict(
        submission="foo.ipynb",
        **run.__kwdefaults__,
    )

    result = run_cli([*cmd_start])
    assert_cli_result(result, expect_error=False)
    mocked_run.assert_called_with(**std_kwargs)

    open("foo.zip", "wb+").close()
    result = run_cli([*cmd_start, "-a", "foo.zip"])
    assert_cli_result(result, expect_error=False)
    mocked_run.assert_called_with(**{**std_kwargs, "autograder": "foo.zip"})

    result = run_cli([*cmd_start, "--autograder", "foo.zip"])
    assert_cli_result(result, expect_error=False)
    mocked_run.assert_called_with(**{**std_kwargs, "autograder": "foo.zip"})

    os.mkdir("out")
    result = run_cli([*cmd_start, "-o", "out"])
    assert_cli_result(result, expect_error=False)
    mocked_run.assert_called_with(**{**std_kwargs, "output_dir": "out"})

    result = run_cli([*cmd_start, "--output-dir", "out"])
    assert_cli_result(result, expect_error=False)
    mocked_run.assert_called_with(**{**std_kwargs, "output_dir": "out"})

    result = run_cli([*cmd_start, "--no-logo"])
    assert_cli_result(result, expect_error=False)
    mocked_run.assert_called_with(**{**std_kwargs, "no_logo": True})

    result = run_cli([*cmd_start, "--debug"])
    assert_cli_result(result, expect_error=False)
    mocked_run.assert_called_with(**{**std_kwargs, "debug": True})

    # test invalid calls
    mocked_run.reset_mock()

    result = run_cli([*cmd_start, "-a", "bar.zip"])
    assert_cli_result(result, expect_error=True)
    mocked_run.assert_not_called()

    result = run_cli([*cmd_start, "-o", "bar"])
    assert_cli_result(result, expect_error=True)
    mocked_run.assert_not_called()



"""Docker container management for Otter Grade"""

import json
import os
import pandas as pd
import pathlib
import pkg_resources
import shutil
import tempfile
import zipfile

from concurrent.futures import ThreadPoolExecutor, wait
from python_on_whales import docker
from textwrap import indent
from typing import List, Optional

from .utils import OTTER_DOCKER_IMAGE_NAME

from ..run.run_autograder.autograder_config import AutograderConfig
from ..utils import loggers


LOGGER = loggers.get_logger(__name__)


def build_image(ag_zip_path: str, base_image: str, tag: str, config: AutograderConfig):
    """
    Creates a grading image based on the autograder zip file and attaches a tag.

    Args:
        ag_zip_path (``str``): path to the autograder zip file
        base_image (``str``): base Docker image to build from
        tag (``str``): tag to be added when creating the image
        config (``otter.run.run_autograder.autograder_config.AutograderConfig``): config overrides
            for the autograder

    Returns:
        ``str``: the tag of the newly-build Docker image
    """
    image = OTTER_DOCKER_IMAGE_NAME + ":" + tag
    dockerfile_path = pkg_resources.resource_filename(__name__, "Dockerfile")

    LOGGER.info(f"Building image using {base_image} as base image")

    with tempfile.TemporaryDirectory() as temp_dir:
        with zipfile.ZipFile(ag_zip_path, 'r') as zip_ref:
            zip_ref.extractall(temp_dir)

        # Update the otter_config.json file from the autograder zip with the provided config
        # overrides.
        config_path = pathlib.Path(temp_dir) / "otter_config.json"
        old_config = AutograderConfig()
        if config_path.exists():
            old_config = AutograderConfig(json.loads(config_path.read_text("utf-8")))

        old_config.update(config.get_user_config())
        config_path.write_text(json.dumps(old_config.get_user_config()))

        docker.build(
            temp_dir,
            build_args={"BASE_IMAGE": base_image},
            tags=[image],
            file=dockerfile_path,
            load=True,
        )

    return image


def launch_containers(
    ag_zip_path: str,
    submission_paths: List[str],
    num_containers: int,
    base_image: str,
    tag: str,
    config: AutograderConfig,
    **kwargs,
):
    """
    Grade submissions in parallel Docker containers.

    This function runs ``num_containers`` Docker containers in parallel to grade the student
    submissions in ``submissions_dir`` using the autograder configuration file at ``ag_zip_path``. 
    If indicated, it copies the PDFs generated of the submissions out of their containers.

    Args:
        ag_zip_path (``str``): path to zip file used to set up container
        submission_paths (``str``): paths of submissions to be graded
        num_containers (``int``): number of containers to run in parallel
        base_image (``str``): the name of a base image to use for building Docker images
        tag (``str``): a tag to use for the ``otter-grade`` image created for this assignment
        config (``otter.run.run_autograder.autograder_config.AutograderConfig``): config overrides
            for the autograder
        **kwargs: additional kwargs passed to ``grade_submission``

    Returns:
        ``list[pandas.core.frame.DataFrame]``: the grades returned by each container spawned
            during grading
    """
    pool = ThreadPoolExecutor(num_containers)
    futures = []
    image = build_image(ag_zip_path, base_image, tag, config)

    for subm_path in submission_paths:
        futures += [pool.submit(
            grade_submission,
            submission_path=subm_path,
            image=image,
            # config=config,
            **kwargs,
        )]

    # stop execution while containers are running
    finished_futures = wait(futures)

    # return list of dataframes
    return [df.result() for df in finished_futures[0]]


def grade_submission(
    submission_path: str,
    image: str,
    no_kill: bool = False,
    pdf_dir: Optional[str] = None,
    timeout: Optional[int] = None,
    network: bool = True,
):
    """
    Grade a submission in a Docker container.

    Args:
        submission_path (``str``): path to the submission to be graded
        image (``str``): a Docker image tag to be used for grading environment
        no_kill (``bool``): whether the grading containers should be kept running after
            grading finishes
        pdf_dir (``str``, optional): a directory in which to put the notebook PDF, if applicable
        timeout (``int``, optional): timeout in seconds for each container
        network (``bool``): whether to enable networking in the containers

    Returns:
        ``pandas.core.frame.DataFrame``: A dataframe of file to grades information
    """
    import dill

    temp_subm_file, temp_subm_path = tempfile.mkstemp()
    shutil.copyfile(submission_path, temp_subm_path)

    results_file, results_path = tempfile.mkstemp(suffix=".pkl")
    pdf_path = None
    if pdf_dir:
        pdf_file, pdf_path = tempfile.mkstemp(suffix=".pdf")

    try:
        nb_basename = os.path.basename(submission_path)
        nb_name = os.path.splitext(nb_basename)[0]

        volumes = [
            (temp_subm_path, f"/autograder/submission/{nb_basename}"),
            (results_path, "/autograder/results/results.pkl")
        ]
        if pdf_dir:
            volumes.append((pdf_path, f"/autograder/submission/{nb_name}.pdf"))

        args = {}
        if network is not None and not network:
            args['networks'] = 'none'

        container = docker.container.create(image, command=["/autograder/run_autograder"], **args)

        for local_path, container_path in volumes:
            docker.container.copy(local_path, (container, container_path))

        docker.container.start(container)

        if timeout:
            import threading

            def kill_container():
                docker.container.kill(container)

            timer = threading.Timer(timeout, kill_container)
            timer.start()

        container_id = container.id[:12]
        LOGGER.info(f"Grading {submission_path} in container {container_id}...")

        exit = docker.container.wait(container)

        if timeout:
            timer.cancel()

        logs = docker.container.logs(container)
        LOGGER.debug(f"Container {container_id} logs:\n{indent(logs, '    ')}")

        # Close our file handles since docker cp will delete the original file when performing the
        # copy.
        os.close(temp_subm_file)
        os.close(results_file)
        if pdf_path:
            os.close(pdf_file)

        for local_path, container_path in volumes:
            docker.container.copy((container, container_path), local_path)

        if not no_kill:
            container.remove()

        if exit != 0:
            raise Exception(
                f"Executing '{submission_path}' in docker container failed! Exit code: {exit}")

        with open(results_path, "rb") as f:
            scores = dill.load(f)

        scores_dict = scores.to_dict()
        scores_dict["percent_correct"] = scores.total / scores.possible

        scores_dict = {t: [scores_dict[t]["score"]] if type(scores_dict[t]) == dict else scores_dict[t] for t in scores_dict}
        scores_dict["file"] = submission_path
        df = pd.DataFrame(scores_dict)

        if pdf_dir:
            os.makedirs(pdf_dir, exist_ok=True)

            local_pdf_path = os.path.join(pdf_dir, f"{nb_name}.pdf")
            shutil.copy(pdf_path, local_pdf_path)

    finally:
        os.remove(results_path)
        os.remove(temp_subm_path)

    return df



"""PDF via LaTeX exporter"""

import nbconvert
import os
import warnings

from textwrap import indent

from .base_exporter import BaseExporter, ExportFailedException, NBCONVERT_6, TEMPLATE_DIR

from ...utils import print_full_width


class PDFViaLatexExporter(BaseExporter):
    """
    Exports notebooks to PDF files using LaTeX as an intermediary

    Converts IPython notebooks to PDFs by first converting them into temporary TeX files that are then
    converted to PDFs using nbconvert and pandoc. Pagebreaks, if enabled, are enforced with a custom
    LaTeX template that clears the document to the next odd numbered page, resulting in responses that
    are all two pages long.

    Attributes:
        default_options (``dict``): the default options for this exporter
    """

    default_options = BaseExporter.default_options.copy()
    default_options.update({
        "save_tex": False,
        "template": "via_latex",
    })

    @classmethod
    def convert_notebook(cls, nb_path, dest, xecjk=False, **kwargs):
        warnings.filterwarnings("ignore", r"invalid escape sequence '\\c'", DeprecationWarning)

        options = cls.default_options.copy()
        options.update(kwargs)

        # choose the template to allow Chinese, Japanese, and Korean characters if necessary
        if xecjk:
            options["template"] = "via_latex_xecjk"

        nb = cls.load_notebook(nb_path, filtering=options["filtering"], pagebreaks=options["pagebreaks"])

        if NBCONVERT_6:
            nbconvert.TemplateExporter.extra_template_basedirs = [TEMPLATE_DIR]
            orig_template_name = nbconvert.TemplateExporter.template_name
            nbconvert.TemplateExporter.template_name = options["template"]

        if options["save_tex"]:
            latex_exporter = nbconvert.LatexExporter()
            if not NBCONVERT_6:
                latex_exporter.template_file = os.path.join(TEMPLATE_DIR, options["template"] + ".tpl")

        pdf_exporter = nbconvert.PDFExporter()
        if not NBCONVERT_6:
            pdf_exporter.template_file = os.path.join(TEMPLATE_DIR, options["template"] + ".tpl")

        try:
            if options["save_tex"]:
                latex_output = nbconvert.export(latex_exporter, nb)
                with open(os.path.splitext(dest)[0] + ".tex", "w+") as output_file:
                    output_file.write(latex_output[0])

            pdf_output = nbconvert.export(pdf_exporter, nb)
            with open(dest, "wb") as output_file:
                output_file.write(pdf_output[0])

        except nbconvert.pdf.LatexFailed as error:
            message = "There was an error generating your LaTeX; showing full error message:\n"
            message += indent(error.output, "    ")
            if xecjk:
                message += "\n\nIf the error above is related to xeCJK or fandol in LaTeX " \
                    "and you don't require this functionality, try running again without " \
                    "xecjk set to True or the --xecjk flag."
            raise ExportFailedException(message)

        finally:
            if NBCONVERT_6:
                nbconvert.TemplateExporter.template_name = orig_template_name



"""Feature toggling for Otter Assign"""

from enum import Enum

from typing import Callable

from .assignment import Assignment


_NOT_RMD: Callable[[Assignment], bool] = lambda a: not a.is_rmd
"""a function that returns ``True`` if the assignment is not an R Markdown assignment"""


# TODO: move other things in
class FeatureToggle(Enum):
    """
    An enum for managing features that are enabled or disabled depending on the assignment config.
    """

    class FeatureToggleValue:

        test: Callable[[Assignment], bool]

        def __init__(self, test: Callable[[Assignment], bool]):
            self.test = test

        def is_enabled(self, assignment: Assignment):
            return self.test(assignment)

    PDF_FILTERING_COMMENTS = FeatureToggleValue(_NOT_RMD)
    """whether PDF filtering HTML comments should be inserted into the notebook"""

    EMPTY_MD_BOUNDARY_CELLS = FeatureToggleValue(_NOT_RMD)
    """whether to use empty Markdown cells for boundaries"""



"""Utilities for Otter Check"""

import ipylab
import nbformat as nbf
import os
import sys
import tempfile
import time
import wrapt

from enum import Enum
from glob import glob
from IPython import get_ipython
from IPython.display import display, Javascript
from ipywidgets import Button, HTML, Output, VBox
from subprocess import run, PIPE

from .logs import EventType

from ..utils import import_or_raise, NBFORMAT_VERSION, NOTEBOOK_METADATA_KEY


def save_notebook(filename, timeout=10):
    """
    Force-saves a Jupyter notebook by displaying JavaScript.

    Args:
        filename (``str``): path to notebook file being saved
        timeout (``int`` or ``float``): number of seconds to wait for save before timing-out

    Returns
        ``bool``: whether the notebook was saved successfully
    """
    if get_ipython() is not None:
        orig_mod_time = os.path.getmtime(filename)
        start = time.time()

        # For classic notebooks
        display(Javascript("""
            if (typeof Jupyter !== 'undefined') {
                Jupyter.notebook.save_checkpoint();
            }
        """))

        # For JupyterLab
        try:
            app = ipylab.JupyterFrontEnd()
            app.commands.execute("docmanager:save")
        except: pass

        while time.time() - start < timeout:
            curr_mod_time = os.path.getmtime(filename)
            if orig_mod_time < curr_mod_time and os.path.getsize(filename) > 0:
                return True

            time.sleep(0.2)

        return False

    return True


def grade_zip_file(zip_path, nb_arcname, tests_dir):
    """
    Grade a submission zip file in a separate process and return the ``GradingResults`` object.
    """
    dill = import_or_raise("dill")

    results_handle, results_path = tempfile.mkstemp(suffix=".pkl")

    try:
        command = [
            sys.executable, "-m", "otter.check.validate_export",
            "--zip-path", zip_path,
            "--nb-arcname", nb_arcname,
            "--tests-dir", tests_dir,
            "--results-path", results_path,
        ]

        # this environment variable is needed to fix #686
        subprocess_env = {**os.environ, "PYDEVD_DISABLE_FILE_VALIDATION": "1"}

        # run the command
        results = run(command, env=subprocess_env, stdout=PIPE, stderr=PIPE)

        # TODO: remove
        print(results.stdout.decode("utf-8"))
        print(results.stderr.decode("utf-8"))

        if results.stderr:
            raise RuntimeError(results.stderr)

        with open(results_path, "rb") as f:
            results = dill.load(f)

        return results

    finally:
        os.close(results_handle)
        os.remove(results_path)


class IPythonInterpreter(Enum):
    """
    An enum of different types of IPython interpreters.
    """

    class Interpreter:
        """
        A class representing a flavor of IPython interpreter.

        Contains attributes for an importable name substring (used to check which interpreter is
        running) and a display name for error messages and the like.
        """

        def __init__(self, check_strs, display_name):
            self.check_strs = check_strs
            self.display_name = display_name

        def running(self):
            """
            Determine whether this interpreter is currently running by checking the string
            representation of the return value of ``IPython.get_ipython``.

            Returns:
                ``bool``: whether this interpreter is running
            """
            ipython_interp = str(get_ipython())
            return any(c in ipython_interp for c in self.check_strs)

    COLAB = Interpreter(["google.colab"], "Google Colab")
    """the Google Colab interpreter"""

    PYOLITE = Interpreter(["pyolite.", "pyodide_kernel."], "Jupyterlite")
    """the JupyterLite interpreter"""


def incompatible_with(interpreter, throw_error = True):
    """
    Create a decorator indicating that a method is incompatible with a specific interpreter.
    """
    @wrapt.decorator
    def incompatible(wrapped, self, args, kwargs):
        """
        A decorator that raises an error or performs no action (depending on ``throw_error``) if the
        wrapped function is called in an environment running on the specified interpreter.
        """
        if self._interpreter is interpreter:
            if throw_error:
                raise RuntimeError(f"This method is not compatible with {interpreter.value.display_name}")
            else:
                return
        return wrapped(*args, **kwargs)

    return incompatible


@wrapt.decorator
def grading_mode_disabled(wrapped, self, args, kwargs):
    """
    A decorator that returns without calling the wrapped function if the ``Notebook`` grading mode
    is enabled.
    """
    if type(self)._grading_mode:
        return
    return wrapped(*args, **kwargs)


# TODO: use this class to replace tuple return values in logs_event
# class LoggedEventReturnValue:
#     def __init__(self, return_value, results=None, shelve_env=None):
#         self.return_value = return_value
#         self.results = results
#         self.shelve_env = shelve_env


def logs_event(event_type):
    """
    A decorator that ensures each call is logged in the Otter log with type ``event_type``.

    Events logging a ``EventType.CHECK`` should return a 3-tuple of the question name, the
    ``TestFile`` and an environment to shelve. All other methods should just return their 
    default return value, which will be logged.
    """
    @wrapt.decorator        
    def event_logger(wrapped, self, args, kwargs):
        """
        Runs a method, catching any errors and logging the call. Returns the return value
        of the function, unless ``EventType.CHECK`` is used, in which case the return value
        is assumed to be a 3-tuple and the second value in the tuple is returned.
        """
        try:
            if event_type == EventType.CHECK:
                question, results, shelve_env = wrapped(*args, **kwargs)

            else:
                results = wrapped(*args, **kwargs)
                shelve_env = {}
                question = None

        except Exception as e:
            self._log_event(event_type, success=False, error=e)
            raise e

        else:
            self._log_event(event_type, results=results, question=question, shelve_env=shelve_env)

        return results

    return event_logger


def list_test_files(tests_dir):
    """
    Find all of the test files in the specified directory (that is, all ``.py`` files that are not
    named ``__init__.py``) and return their paths in a sorted list.

    Args:
        tests_dir (``str``): the path to the tests directory

    Returns:
        ``list[str]``: the sorted list of all test file paths in ``tests_dir``
    """
    return sorted([file for file in glob(os.path.join(tests_dir, "*.py")) \
            if file != "__init__.py"])


def list_available_tests(tests_dir, nb_path):
    """
    Get a list of available questions by searching the tests directory (if present) or the notebook
    metadata.

    Args:
        tests_dir (``str``): the path to the tests directory
        nb_path (``str``): the path to the notebook

    Returns:
        ``list[str]``: the sorted list of question names
    """
    get_stem = lambda p: os.path.splitext(os.path.basename(p))[0]

    if tests_dir and os.path.isdir(tests_dir):
        tests = map(get_stem, list_test_files(tests_dir))

    else:
        if nb_path is None:
            raise ValueError("Tests directory does not exist and no notebook path provided")

        nb = nbf.read(nb_path, as_version=NBFORMAT_VERSION)
        tests = list(nb["metadata"][NOTEBOOK_METADATA_KEY]["tests"].keys())

    return sorted(tests)


def resolve_test_info(tests_dir, nb_path, tests_url_prefix, question):
    """
    Determine the test path and test name.

    If ``tests_url_prefix`` is specified, the test file is downloaded from the URL
    ``{tests_url_prefix}/{question}.py`` and saved to the file ``{tests_dir}/{question}.py``. If
    ``tests_dir`` does not already exist, it is created.

    Args:
        tests_dir (``str``): the path to the directory of tests
        nb_path (``str``): the path to the notebook
        tests_url_prefix (``str | None``): the prefix of a URL to the test file
        question (``str``): the question name

    Returns:
        ``tuple[str, str]``: the test path and test name
    """
    if tests_url_prefix is not None:
        if not IPythonInterpreter.PYOLITE.value.running():
            raise ValueError("Downloading test files from URLs is only supported on JupyterLite")

        pyodide = import_or_raise("pyodide")

        test_url = f"{tests_url_prefix}{'/' if not tests_url_prefix.endswith('/') else ''}{question}.py"
        text = pyodide.open_url(test_url).getvalue()

        os.makedirs(tests_dir, exist_ok=True)

        test_path = os.path.join(tests_dir, f"{question}.py")
        test_name = None

        with open(test_path, "w+") as f:
            f.write(text)

    elif tests_dir and os.path.isdir(tests_dir):
        if not os.path.isfile(os.path.join(tests_dir, question + ".py")):
            raise FileNotFoundError(f"Test {question} does not exist")

        test_path = os.path.join(tests_dir, question + ".py")
        test_name = None

    else:
        if nb_path is None:
            raise ValueError("Tests directory does not exist and no notebook path provided")

        test_path = nb_path
        test_name = question

    return test_path, test_name


def display_pdf_confirmation_widget(message, callback):
    """
    Display a widget to the user to acknowledge that a PDF will not be included in their submission
    zip.

    Args:
        message (``str | None``): a custom message to use
        callback (``callable[]``): a callback function to execute after the user ACKs
    """
    o = Output()
    def wrapped_callback(*args):
        with o: callback()

    if not message:
        message = "Your notebook could not be exported as a PDF. To continue exporting your " \
            "submission, please click the button below."

    t = HTML(f"""<p style="margin: 0">{message}</p>""")
    b = Button(description="Continue export", button_style="warning")
    b.on_click(wrapped_callback)
    m = HTML("""<div style="height: 10px; width: 100%"></div>""")
    
    display(VBox([t, b, m, o]))



"""Execution and grading internals for Otter-Grader"""

import nbformat
import pickle
import tempfile

from traitlets.config import Config

from .checker import Checker
from .logging import start_server
from .preprocessor import GradingPreprocessor

from ..test_files import GradingResults
from ..utils import NBFORMAT_VERSION


def grade_notebook(
    submission_path,
    *,
    tests_glob=[],
    ignore_errors=True,
    script=False, 
    cwd=None,
    test_dir=None,
    seed=None,
    seed_variable=None,
    log=None,
    variables=None, 
    plugin_collection=None,
):
    """
    Grade an assignment file and return grade information.

    Args:
        submission_path (``str``): path to a single notebook or Python script
        tests_glob (``list[str]``): paths of test files that should be run; tests that are included
            in this list that have not already been run by the submission are run after executing
            the submission
        ignore_errors (``bool``): whether errors in execution should be ignored
        script (``bool``): whether the ``submission_path`` is a Python script
        cwd (``str``): working directory of execution to be appended to ``sys.path`` before
            executing the submission
        test_dir (``str``): path to directory of tests in the grading environment
        seed (``int``): random seed for intercell seeding
        seed_variable (``str|None``): a variable name to override with the seed
        log (``otter.check.logs.Log``): an Otter log from which to deserialize environments to grade
            questions
        variables (``dict[str, str]|None``): map of variable names -> type string to use for type
            checking values deserialized from ``log``
        plugin_collection (``otter.plugins.PluginCollection``): a set of plugins to run the
            ``before_execution`` and ``after_grading`` events on this submission

    Returns:
        ``otter.test_files.GradingResults``: the results of grading
    """
    from nbconvert.preprocessors import ExecutePreprocessor

    if not script:
        nb = nbformat.read(submission_path, as_version=NBFORMAT_VERSION)

    else:
        with open(submission_path) as f:
            nb = f.read()

        nb = nbformat.v4.new_notebook(cells=[nbformat.v4.new_code_cell(nb)])

    if plugin_collection is not None:
        nb = plugin_collection.before_execution(nb)

    with tempfile.NamedTemporaryFile() as ntf:
        c = Config()

        (host, port), stop_server = start_server()

        # GradingPreprocessor config
        c.GradingPreprocessor.cwd = cwd
        c.GradingPreprocessor.test_dir = test_dir
        c.GradingPreprocessor.tests_glob = tests_glob
        c.GradingPreprocessor.results_path = ntf.name
        c.GradingPreprocessor.seed = seed
        c.GradingPreprocessor.seed_variable = seed_variable
        c.GradingPreprocessor.otter_log = log
        c.GradingPreprocessor.variables = variables
        c.GradingPreprocessor.logging_server_host = host
        c.GradingPreprocessor.logging_server_port = port

        # ExecutePreprocessor config
        c.ExecutePreprocessor.allow_errors = ignore_errors

        gp = GradingPreprocessor(config=c)
        ep = ExecutePreprocessor(config=c)

        nb, _ = gp.preprocess(nb)
        executed_nb, _ = ep.preprocess(nb)

        stop_server()

        gp.cleanup()

        results = pickle.load(ntf)

    if not isinstance(results, GradingResults):
        raise TypeError("Results deserialized from grading notebook were not a GradingResults instance")

    results.notebook = executed_nb

    if plugin_collection is not None:
        plugin_collection.run("after_grading", results)

    return results



"""Tests for ``otter.check``"""

import contextlib
import os
import pytest

from glob import glob
from io import StringIO
from textwrap import dedent
from unittest import mock

from otter.check import main as check

from ..utils import TestFileManager


FILE_MANAGER = TestFileManager(__file__)


@pytest.fixture(autouse=True)
def cleanup_check_output(cleanup_enabled):
    yield
    if cleanup_enabled and os.path.exists(".OTTER_LOG"):
        os.remove(".OTTER_LOG")


def test_otter_check_script():
    """
    Checks that the script checker works
    """
    # run for each individual test
    for file in glob(FILE_MANAGER.get_path("tests/*.py")):
        # capture stdout
        output = StringIO()
        with contextlib.redirect_stdout(output):
            check(
                FILE_MANAGER.get_path("file0.py"),
                question = os.path.split(file)[1][:-3],
                tests_path = os.path.split(file)[0],
            )

        if os.path.split(file)[1] != "q2.py":
            assert output.getvalue().strip().split("\n")[-1].strip() == \
                "All tests passed!", \
                "Did not pass test at {}".format(file)

    # run the file for all questions
    output = StringIO()
    with contextlib.redirect_stdout(output):
        check(
            FILE_MANAGER.get_path("file0.py"), 
            tests_path = os.path.split(file)[0],
        )

    assert output.getvalue().strip() == \
        dedent("""\
            q1 results: All test cases passed!
            q2 results:
                q2 - 1 result:
                    ❌ Test case failed
                    Trying:
                        1 == 1
                    Expecting:
                        False
                    **********************************************************************
                    Line 2, in q2 0
                    Failed example:
                        1 == 1
                    Expected:
                        False
                    Got:
                        True

                q2 - 2 result:
                    ✅ Test case passed
            q3 results: All test cases passed!
            q4 results: All test cases passed!
            q5 results: All test cases passed!"""), \
        "Did not pass correct tests"


def test_otter_check_notebook():
    """
    Checks that the script checker works
    """
    # run for each individual test
    for file in glob(FILE_MANAGER.get_path("tests/*.py")):
        # capture stdout
        output = StringIO()
        with contextlib.redirect_stdout(output):
            check(
                FILE_MANAGER.get_path("test-nb.ipynb"), 
                question = os.path.split(file)[1][:-3],
                tests_path = os.path.split(file)[0],
            )

        if os.path.split(file)[1] != "q2.py":
            assert output.getvalue().strip().split("\n")[-1].strip() == \
                "All tests passed!", \
                "Did not pass test at {}".format(file)

    # run the file for all questions
    output = StringIO()
    with contextlib.redirect_stdout(output):
        check(
            FILE_MANAGER.get_path("test-nb.ipynb"), 
            tests_path = os.path.split(file)[0],
        )

    assert output.getvalue().strip() == \
        dedent("""\
            q1 results: All test cases passed!
            q2 results:
                q2 - 1 result:
                    ❌ Test case failed
                    Trying:
                        1 == 1
                    Expecting:
                        False
                    **********************************************************************
                    Line 2, in q2 0
                    Failed example:
                        1 == 1
                    Expected:
                        False
                    Got:
                        True

                q2 - 2 result:
                    ✅ Test case passed
            q3 results: All test cases passed!
            q4 results: All test cases passed!
            q5 results: All test cases passed!"""), \
        "Did not pass correct tests"



"""IPython notebook API for Otter Check"""

import datetime as dt
import inspect
import json
import os
import warnings
import zipfile

from glob import glob
from IPython.display import display, HTML
from textwrap import indent

from .logs import LogEntry, EventType, Log
from .utils import (
    display_pdf_confirmation_widget,
    grade_zip_file,
    grading_mode_disabled,
    incompatible_with,
    IPythonInterpreter,
    list_available_tests,
    logs_event,
    resolve_test_info,
    save_notebook,
)

from ..execute import Checker
from ..export import export_notebook
from ..plugins import PluginCollection
from ..test_files import GradingResults
from ..utils import (
    Loggable,
    NO_PDF_EXPORT_MESSAGE_KEY,
    NOTEBOOK_METADATA_KEY,
    REQUIRE_CONFIRMATION_NO_PDF_EXPORT_KEY,
)


_OTTER_LOG_FILENAME = ".OTTER_LOG"
_SHELVE = False
_ZIP_NAME_FILENAME = "__zip_filename__"


class Notebook(Loggable):
    """
    Notebook class for in-notebook autograding

    Args:
        nb_path(``str``, optional): path to the notebook being run
        tests_dir (``str``, optional): path to tests directory
        colab (``bool``, optional): whether this notebook is being run on Google Colab; if ``None``,
            this information is automatically parsed from IPython on creation
        jupyterlite (``bool``, optional): whether this notebook is being run on JupyterLite; if
            ``None``, this information is automatically parsed from IPython on creation
    """

    _grading_mode = False

    # overrides tests_dir arg in __init__, used for changing tests dir during grading
    _tests_dir_override = None

    @logs_event(EventType.INIT)
    def __init__(
        self,
        nb_path=None,
        tests_dir="./tests",
        tests_url_prefix=None,
        colab=None,
        jupyterlite=None,
    ):
        global _SHELVE

        interpreter = None
        if colab or IPythonInterpreter.COLAB.value.running():
            interpreter = IPythonInterpreter.COLAB
        elif jupyterlite or IPythonInterpreter.PYOLITE.value.running():
            interpreter = IPythonInterpreter.PYOLITE

        self._interpreter = interpreter

        if self._interpreter is IPythonInterpreter.COLAB and not os.path.isdir(tests_dir):
            raise ValueError(f"Tests directory {tests_dir} does not exist")

        if type(self)._tests_dir_override is not None:
            self._path = type(self)._tests_dir_override
        else:
            self._path = tests_dir

        # TODO: clean up logic for handling nb path; this can be determined in __init__ instead of
        # begin computed every time. Maybe require it as an argument so it can be removed from other
        # methods?
        self._notebook = nb_path
        self._tests_url_prefix = tests_url_prefix
        self._addl_files = []
        self._plugin_collections = {}

        # assume using otter service if there is a .otter file
        otter_configs = glob("*.otter")
        if otter_configs:
            # check that there is only 1 config
            assert len(otter_configs) == 1, "More than 1 otter config file found"

            # load in config file
            with open(otter_configs[0], encoding="utf-8") as f:
                self._config = json.load(f)

            _SHELVE = self._config.get("save_environment", False)
            self._ignore_modules = self._config.get("ignore_modules", [])
            self._vars_to_store = self._config.get("variables", None)

            self._notebook = self._config["notebook"]

    @classmethod
    def init_grading_mode(cls, tests_dir):
        logger = cls._get_logger()
        logger.info("Entering Notebook grading mode")
        logger.debug(f"Overriding tests directory: {tests_dir}")
        cls._grading_mode = True
        cls._tests_dir_override = tests_dir
        Checker.clear_results()
        Checker.enable_tracking()

    @incompatible_with(IPythonInterpreter.PYOLITE, throw_error=False)
    def _log_event(self, event_type, results=[], question=None, success=True, error=None, shelve_env={}):
        """
        Logs an event

        Args:
            event_type (``otter.logs.EventType``): the type of event
            results (``list[otter.test_files.abstract_test.TestFile]``, optional):
                the results of any checks recorded by the entry
            question (``str``, optional): the question name for this check
            success (``bool``, optional): whether the operation was successful
            error (``Exception``, optional): the exception thrown by the operation, if applicable
        """
        entry = LogEntry(
            event_type,
            results=results,
            question=question,
            success=success,
            error=error
        )

        if _SHELVE and event_type == EventType.CHECK:
            entry.shelve(
                shelve_env,
                delete=True,
                filename=_OTTER_LOG_FILENAME,
                ignore_modules=self._ignore_modules,
                variables=self._vars_to_store
            )

        entry.flush_to_file(_OTTER_LOG_FILENAME)

    def _resolve_nb_path(self, nb_path, fail_silently=False):
        """
        Attempts to resolve the path to the notebook being run. If ``nb_path`` is ``None``, ``self._notebook``
        is checked, then the working directory is searched for ``.ipynb`` files.

        Args:
            nb_path (``Optional[str]``): path to the notebook
            fail_silently (``bool``): if true, the method does not fail the notebook path can't be
                resolved

        Returns:
            ``str``: resolved notebook path

        Raises:
            ``ValueError``: if no notebooks or too many notebooks are found.
        """
        if nb_path is None and self._notebook is not None:
            nb_path = self._notebook
            if not os.path.isfile(nb_path):
                raise ValueError(f"Expected a notebook file named '{nb_path}' but no such file found")

        elif nb_path is None and glob("*.ipynb"):
            notebooks = glob("*.ipynb")
            if len(notebooks) == 1:
                nb_path = notebooks[0]

        if nb_path is None and not fail_silently:
            raise ValueError("Could not resolve notebook path")

        return nb_path

    @logs_event(EventType.CHECK)
    def check(self, question, global_env=None):
        """
        Runs tests for a specific question against a global environment. If no global environment
        is provided, the test is run against the calling frame's environment.

        Args:
            question (``str``): name of question being graded
            global_env (``dict``, optional): global environment resulting from execution of a single
                notebook

        Returns:
            ``otter.test_files.abstract_test.TestFile``: the grade for the question
        """
        self._logger.info(f"Running check for question: {question}")
        test_path, test_name = resolve_test_info(
            self._path,
            self._resolve_nb_path(None, fail_silently=True),
            self._tests_url_prefix,
            question,
        )

        self._logger.debug(f"Resolved test path: {test_path}")
        self._logger.debug(f"Resolved test name: {test_name}")

        # raise an error for a metadata test on Colab
        if test_name is not None and self._interpreter is IPythonInterpreter.COLAB:
            raise ValueError(f"Test {question} does not exist")

        # ensure that desired test exists
        if not os.path.isfile(test_path):
            raise FileNotFoundError(f"Test {question} does not exist")

        # pass the correct global environment
        if global_env is None:
            self._logger.debug(f"Collecting calling global environment")
            global_env = inspect.currentframe().f_back.f_back.f_globals

        # run the check
        self._logger.debug(f"Calling checker")
        result = Checker.check(test_path, test_name, global_env)

        return question, result, global_env

    @incompatible_with(IPythonInterpreter.COLAB)
    def run_plugin(self, plugin_name, *args, nb_path=None, **kwargs):
        """
        Runs the plugin ``plugin_name`` with the specified arguments. Use ``nb_path`` if the path
        to the notebook is not configured.

        Args:
            plugin_name (``str``): importable name of an Otter plugin that implements the 
                ``from_notebook`` hook
            *args: arguments to be passed to the plugin
            nb_path (``str``, optional): path to the notebook
            **kwargs: keyword arguments to be passed to the plugin

        """
        self._logger.info(f"Running plugin {plugin_name}")
        self._logger.debug(f"Running plugin {plugin_name} with args {args} and kwargs {kwargs}")
        nb_path = self._resolve_nb_path(nb_path)
        if plugin_name in self._plugin_collections:
            pc = self._plugin_collections[plugin_name]
        else:
            pc = PluginCollection([plugin_name], nb_path, {})
            self._plugin_collections[plugin_name] = pc
        pc.run("from_notebook", *args, **kwargs)

    @grading_mode_disabled
    @incompatible_with(IPythonInterpreter.COLAB)
    @logs_event(EventType.TO_PDF)
    def to_pdf(self, nb_path=None, filtering=True, pagebreaks=True, display_link=True, force_save=False):
        """
        Exports a notebook to a PDF using Otter Export

        Args:
            nb_path (``str``, optional): path to the notebook we want to export; will attempt to infer
                if not provided
            filtering (``bool``, optional): set true if only exporting a subset of notebook cells to PDF
            pagebreaks (``bool``, optional): if true, pagebreaks are included between questions
            display_link (``bool``, optional): whether or not to display a download link
            force_save (``bool``, optional): whether or not to display JavaScript that force-saves the
                notebook (only works in Jupyter Notebook classic, not JupyterLab)
        """
        nb_path = self._resolve_nb_path(nb_path)
        self._logger.debug(f"Resolved notebook path: {nb_path}")

        if force_save:
            self._logger.debug("Attempting to force-save notebook")
            saved = save_notebook(nb_path)
            if not saved:
                warnings.warn(
                    "Couldn't automatically save the notebook; we recommend using File > Save & "
                    "Checkpoint and then re-running this cell. The zip file returned by this call "
                    "will use the last saved version of this notebook."
                )
            else:
                self._logger.debug("Force-save successful")

        pdf_path = export_notebook(nb_path, filtering=filtering, pagebreaks=pagebreaks)
        self._logger.debug(f"Wrote PDF to zip file: {pdf_path}")

        if display_link:
            # create and display output HTML
            out_html = f"""
            <p>Your file has been exported. Download it by right-clicking
            <a href="{pdf_path}" target="_blank">here</a> and selecting <strong>Save Link As</strong>.
            """

            display(HTML(out_html))

    @grading_mode_disabled
    @incompatible_with(IPythonInterpreter.COLAB)
    def add_plugin_files(self, plugin_name, *args, nb_path=None, **kwargs):
        """
        Runs the ``notebook_export`` event of the plugin ``plugin_name`` and tracks the file paths
        it returns to be included when calling ``Notebook.export``.

        Args:
            plugin_name (``str``): importable name of an Otter plugin that implements the 
                ``from_notebook`` hook
            *args: arguments to be passed to the plugin
            nb_path (``str``, optional): path to the notebook
            **kwargs: keyword arguments to be passed to the plugin        
        """
        nb_path = self._resolve_nb_path(nb_path)
        if plugin_name in self._plugin_collections:
            pc = self._plugin_collections[plugin_name]
        else:
            pc = PluginCollection([plugin_name], nb_path, {})
            self._plugin_collections[plugin_name] = pc
        addl_files = pc.run("notebook_export", *args, **kwargs)[0]
        if addl_files is None:
            return
        self._addl_files.extend(addl_files)

    @grading_mode_disabled
    @incompatible_with(IPythonInterpreter.COLAB)
    @logs_event(EventType.END_EXPORT)
    def export(self, nb_path=None, export_path=None, pdf=True, filtering=True, pagebreaks=True, files=[], 
            display_link=True, force_save=False, run_tests=False):
        """
        Exports a submission to a zip file. Creates a submission zipfile from a notebook at ``nb_path``,
        optionally including a PDF export of the notebook and any files in ``files``.

        Args:
            nb_path (``str``, optional): path to the notebook we want to export; will attempt to infer
                if not provided
            export_path (``str``, optional): path at which to write zipfile; defaults to notebook's
                name + ``.zip``
            pdf (``bool``, optional): whether a PDF should be included
            filtering (``bool``, optional): whether the PDF should be filtered; ignored if ``pdf`` is
                ``False``
            pagebreaks (``bool``, optional): whether pagebreaks should be included between questions
                in the PDF
            files (``list`` of ``str``, optional): paths to other files to include in the zip file
            display_link (``bool``, optional): whether or not to display a download link
            force_save (``bool``, optional): whether or not to display JavaScript that force-saves the
                notebook (only works in Jupyter Notebook classic, not JupyterLab)
        """
        self._log_event(EventType.BEGIN_EXPORT)

        nb_path = self._resolve_nb_path(nb_path)
        self._logger.debug(f"Resolved notebook path: {nb_path}")

        if force_save:
            self._logger.debug("Attempting to force-save notebook")
            saved = save_notebook(nb_path)
            if not saved:
                warnings.warn(
                    "Couldn't automatically save the notebook; we recommend using File > Save & "
                    "Checkpoint and then re-running this cell. The zip file returned by this call "
                    "will use the last saved version of this notebook."
                )
            else:
                self._logger.debug("Force-save successful")

        otter_nb_config = {}
        with open(nb_path, "r", encoding="utf-8") as f:
            if len(f.read().strip()) == 0:
                raise ValueError(f"Notebook '{nb_path}' is empty. Please save and checkpoint your "
                    "notebook and rerun this cell.")

            f.seek(0)
            # TODO: turn the notebook metadata configs into a fica.Config
            otter_nb_config = json.load(f).get("metadata", {}).get(NOTEBOOK_METADATA_KEY, {})

        timestamp = dt.datetime.now().strftime("%Y_%m_%dT%H_%M_%S_%f")
        if export_path is None:
            zip_path = ".".join(nb_path.split(".")[:-1]) + "_" + timestamp + ".zip"
        else:
            zip_path = export_path

        self._logger.debug(f"Determined export zip path: {zip_path}")

        zf = zipfile.ZipFile(zip_path, mode="w")
        zf.write(nb_path)

        pdf_created = True
        if pdf:
            pdf_path = export_notebook(nb_path, filtering=filtering, pagebreaks=pagebreaks)
            if os.path.isfile(pdf_path):
                pdf_created = True
                zf.write(pdf_path)
                self._logger.debug(f"Wrote PDF to zip file: {pdf_path}")
            else:
                pdf_created = False
                warnings.warn("Could not locate a PDF to include")

        def continue_export():
            if os.path.isfile(_OTTER_LOG_FILENAME):
                zf.write(_OTTER_LOG_FILENAME)
                self._logger.debug("Added Otter log to zip file")

            zip_basename = os.path.basename(zip_path)
            zf.writestr(_ZIP_NAME_FILENAME, zip_basename)
            self._logger.debug(f"Added {_ZIP_NAME_FILENAME} to zip file: '{zip_basename}'")

            dot_otter = glob("*.otter")
            if dot_otter:
                if len(dot_otter) != 1:
                    raise ValueError("Too many .otter files (max 1 allowed)")
                dot_otter = dot_otter[0]
                zf.write(dot_otter)
                self._logger.debug(f"Added .otter file to zip file: {dot_otter}")

            for file in files:
                if os.path.isdir(file):
                    sub_files = glob(f"{file}/**/*.*", recursive=True)
                    for sub_file in sub_files:
                        zf.write(sub_file)
                else:
                    zf.write(file)
                self._logger.debug(f"Added file to zip file: {file}")

            for file in self._addl_files:
                zf.write(file)
                self._logger.debug(f"Added plugin file to zip file: {file}")

            zf.close()

            if run_tests:
                print("Running your submission against local test cases...\n")
                results = grade_zip_file(zip_path, nb_path, self._path)
                print(
                    "Your submission received the following results when run against " + \
                    "available test cases:\n\n" + indent(results.summary(), "    "))

            if display_link:
                # create and display output HTML
                out_html = f"""
                    <p>
                        Your submission has been exported. Click
                        <a href="{zip_path}" downloadzip_path target="_blank">here</a> to download
                        the zip file.
                    </p>
                """

                display(HTML(out_html))

        if pdf_created or not otter_nb_config.get(REQUIRE_CONFIRMATION_NO_PDF_EXPORT_KEY, False):
            continue_export()
        else:
            display_pdf_confirmation_widget(
                otter_nb_config.get(NO_PDF_EXPORT_MESSAGE_KEY), continue_export)

    @grading_mode_disabled
    @logs_event(EventType.END_CHECK_ALL)
    def check_all(self):
        """
        Runs all tests on this notebook. Tests are run against the current global environment, so any
        tests with variable name collisions will fail.
        """
        self._log_event(EventType.BEGIN_CHECK_ALL)

        tests = list_available_tests(self._path, self._resolve_nb_path(None, fail_silently=True))

        global_env = inspect.currentframe().f_back.f_back.f_back.f_globals

        self._logger.debug(f"Found available tests: {', '.join(tests)}")

        results = []
        if not _SHELVE:
            for test_name in tests:
                results.append(self.check(test_name, global_env))

        else:
            log = Log.from_file(_OTTER_LOG_FILENAME, ascending=False)
            for file in sorted(tests):
                if "__init__.py" not in file:
                    test_name = os.path.splitext(os.path.split(file)[1])[0]

                    entry = log.get_question_entry(test_name)
                    env = entry.unshelve()
                    global_env.update(env)
                    del locals()["env"]

                    result = self.check(test_name, global_env)
                    results.append((test_name, result))

        return GradingResults(results)



"""Tests for ``otter.generate``"""

import os
import pytest

from unittest import mock

from otter.generate import main as generate

from ..utils import assert_dirs_equal, TestFileManager, unzip_to_temp


FILE_MANAGER = TestFileManager(__file__)
OUTPUT_PATH = FILE_MANAGER.get_path("autograder.zip")


@pytest.fixture(autouse=True)
def cleanup_output(cleanup_enabled):
    yield
    if cleanup_enabled and os.path.exists(OUTPUT_PATH):
        os.remove(OUTPUT_PATH)


def test_autograder():
    """
    Check that the correct zip file is generated.
    """
    # create the zipfile
    generate(
        tests_dir = FILE_MANAGER.get_path("tests"),
        output_path = OUTPUT_PATH,
        requirements = FILE_MANAGER.get_path("requirements.txt"),
        files = [FILE_MANAGER.get_path("data/test-df.csv")],
        no_environment = True,  # don't use the environment.yml in the root of the repo
    )

    with unzip_to_temp(FILE_MANAGER.get_path("autograder.zip")) as unzipped_dir:
        assert_dirs_equal(unzipped_dir, FILE_MANAGER.get_path("autograder-correct"))


@mock.patch("otter.generate.APIClient")
def test_autograder_with_token(mocked_client):
    """
    Check that the correct zip file is generated when a token is specified instead of a username
    and password.

    This test also checks that directories are copied correctly when listed in the ``files``
    argument.
    """
    # create the zipfile
    generate(
        tests_dir = FILE_MANAGER.get_path("tests"),
        output_path = OUTPUT_PATH,
        requirements = FILE_MANAGER.get_path("requirements.txt"),
        config = FILE_MANAGER.get_path("otter_config.json"),
        files = [FILE_MANAGER.get_path("data")],
        no_environment = True,  # don't use the environment.yml in the root of the repo
    )

    mocked_client.assert_not_called()

    with unzip_to_temp(FILE_MANAGER.get_path("autograder.zip")) as unzipped_dir:
        assert_dirs_equal(unzipped_dir, FILE_MANAGER.get_path("autograder-token-correct"))


def test_custom_env():
    """
    Check that a custom environment.yml is correctly read and modified.
    """
    # create the zipfile
    generate(
        tests_dir = FILE_MANAGER.get_path("tests"),
        output_path = OUTPUT_PATH,
        requirements = FILE_MANAGER.get_path("requirements.txt"),
        environment = FILE_MANAGER.get_path("environment.yml"),
        files = [FILE_MANAGER.get_path("data/test-df.csv")],
    )

    with unzip_to_temp(FILE_MANAGER.get_path("autograder.zip")) as unzipped_dir:
        assert_dirs_equal(unzipped_dir, FILE_MANAGER.get_path("autograder-custom-env"))


def test_lang_r():
    """
    Check that the R autograder configuration is generated correctly.
    """
    # create the zipfile
    generate(
        tests_dir = FILE_MANAGER.get_path("tests"),
        output_path = OUTPUT_PATH,
        config = FILE_MANAGER.get_path("r_otter_config.json"),
        no_environment = True,
        channel_priority_strict = False,
    )

    with unzip_to_temp(FILE_MANAGER.get_path("autograder.zip")) as unzipped_dir:
        assert_dirs_equal(unzipped_dir, FILE_MANAGER.get_path("autograder-r-correct"))


def test_r_with_requirements():
    """
    Check that the R autograder configuration with a requirements file is generated correctly.
    """
    # create the zipfile
    generate(
        tests_dir = FILE_MANAGER.get_path("tests"),
        output_path = OUTPUT_PATH,
        config = FILE_MANAGER.get_path("r_otter_config.json"),
        requirements = FILE_MANAGER.get_path("requirements.r"),
        no_environment = True,
        channel_priority_strict = False,
    )

    with unzip_to_temp(FILE_MANAGER.get_path("autograder.zip")) as unzipped_dir:
        assert_dirs_equal(unzipped_dir, FILE_MANAGER.get_path("autograder-r-requirements-correct"))



"""Abstract test objects for providing a schema to write and parse test cases"""

import random
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass, replace
from textwrap import indent
from typing import Optional, Union


OK_FORMAT_VARNAME = "OK_FORMAT"


# # class for storing the test cases themselves
# #   - body is the string that gets run for the test
# #   - hidden is the visibility of the test case
# #   - points is the number of points this test case is worth
# TestCase = namedtuple("TestCase", ["name", "body", "hidden", "points", "success_message", "failure_message"])

@dataclass
class TestCase:

    name: str

    body: str

    hidden: bool

    points: Union[int, float]

    success_message: Optional[str]

    failure_message: Optional[str]


@dataclass
class TestCaseResult:

    test_case: TestCase

    message: Optional[str]

    passed: bool

# # class for storing the results of a single test _case_ (within a test file)
# #   - message should be a string to print out to the student (ignored if passed is True)
# #   - passed is whether the test case passed
# #   - hidden is the visibility of the test case
# TestCaseResult = namedtuple("TestCaseResult", ["test_case", "message", "passed"])


class TestFile(ABC):
    """
    A (abstract) single test file for Otter. This ABC defines how test results are represented and sets
    up the instance variables tracked by tests. Subclasses should implement the abstract class method
    ``from_file`` and the abstract instance method ``run``.

    Args:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        all_or_nothing (``bool``, optional): whether the test should be graded all-or-nothing across
            cases

    Attributes:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        all_or_nothing (``bool``): whether the test should be graded all-or-nothing across
            cases
        test_case_results (``list`` of ``TestCaseResult``): a list of results for the test cases in
            ``test_cases``
    """

    def _repr_html_(self):
        if self.passed_all:
            all_passed_emoji = random.choice(['🍀', '🎉', '🌈', '🙌', '🚀', '🌟', '✨', '💯'])
            if any(tcr.test_case.success_message is not None for tcr in self.test_case_results):
                ret = f"<p><strong><pre style='display: inline;'>{self.name}</pre></strong> passed! {all_passed_emoji}</p>"
                for tcr in self.test_case_results:
                    if tcr.test_case.success_message is not None:
                        ret += f"<p><strong><pre style='display: inline;'>{tcr.test_case.name}</pre> message:</strong> {tcr.test_case.success_message}</p>"
                return ret
            return f"<p><strong><pre style='display: inline;'>{self.name}</pre></strong> passed! {all_passed_emoji}</p>"
        else:
            ret = f"<p><strong style='color: red;'><pre style='display: inline;'>{self.name}</pre> results:</strong></p>"
            for tcr in self.test_case_results:
                if tcr.passed and tcr.test_case.success_message is not None:
                    ret += f"<p><strong><pre style='display: inline;'>{tcr.test_case.name}</pre> message:</strong> {tcr.test_case.success_message}</p>"
                if not tcr.passed and tcr.test_case.failure_message is not None:
                    ret += f"<p><strong><pre style='display: inline;'>{tcr.test_case.name}</pre> message:</strong> {tcr.test_case.failure_message}</p>"
                ret += f"<p><strong><pre style='display: inline;'>{tcr.test_case.name}</pre> result:</strong></p>"
                ret += f"<pre>{indent(tcr.message, '    ')}</pre>"

            return ret

    def __repr__(self):
        return self.summary()

    # @abstractmethod
    def __init__(self, name, path, test_cases, all_or_nothing=True):
        self.name = name
        self.path = path
        self.test_cases = test_cases
        self.all_or_nothing = all_or_nothing
        self.test_case_results = []
        self._score = None

    @staticmethod
    def resolve_test_file_points(total_points, test_cases):
        if isinstance(total_points, list):
            if len(total_points) != len(test_cases):
                raise ValueError("Points specified in test has different length than number of test cases")
            test_cases = [replace(tc, points=pt) for tc, pt in zip(test_cases, total_points)]
            total_points = None

        elif total_points is not None and not isinstance(total_points, (int, float)):
            raise TypeError(f"Test spec points has invalid type: {total_points}")

        point_values = []
        for i, test_case in enumerate(test_cases):
            if test_case.points is not None:
                assert type(test_case.points) in (int, float), f"Invalid point type: {type(test_case.points)}"
                point_values.append(test_case.points)

            else:
                point_values.append(None)

        pre_specified = sum(p for p in point_values if p is not None)
        if total_points is not None:
            if pre_specified > total_points:
                raise ValueError(f"More points specified in test cases than allowed for test")

            else:
                try:
                    per_remaining = (total_points - pre_specified) / sum(1 for p in point_values if p is None)
                except ZeroDivisionError:
                    per_remaining = 0.0

        else:
            if pre_specified == 0 and all(p in (0, None) for p in point_values):
                # if only zeros specified, assume test worth 1 pt and divide amongst nonzero cases
                try:
                    per_remaining = 1 / sum(p is None for p in point_values)
                except ZeroDivisionError:
                    per_remaining = 0.0

            elif pre_specified == 0:
                per_remaining = 1 / len(point_values)

            else:
                # assume all other tests are worth 0 points
                per_remaining = 0.0

        point_values = [p if p is not None else per_remaining for p in point_values]
        return [replace(tc, points=p) for tc, p in zip(test_cases, point_values)]

    @property
    def passed_all(self):
        return all(tcr.passed for tcr in self.test_case_results)

    @property
    def passed_all_public(self):
        return all(tcr.passed for tcr in self.test_case_results if not tcr.test_case.hidden)

    @property
    def all_public(self):
        return all(not tc.hidden for tc in self.test_cases)

    @property
    def grade(self):
        if self.all_or_nothing and not self.passed_all:
            return 0
        elif self.all_or_nothing and self.passed_all:
            return 1
        else:
            return sum(tcr.test_case.points for tcr in self.test_case_results if tcr.passed) / \
                sum(tc.points for tc in self.test_cases)

    @property
    def score(self):
        if self._score is not None:
            return self._score
        return sum(tcr.test_case.points for tcr in self.test_case_results if tcr.passed)

    @property
    def possible(self):
        return sum(tc.points for tc in self.test_cases)

    def update_score(self, new_score):
        self._score = new_score

    def to_dict(self):
        return {
            "score": self.score,
            "possible": self.possible,
            "name": self.name,
            "path": self.path,
            "test_cases": [asdict(tc) for tc in self.test_cases],
            "all_or_nothing": self.all_or_nothing,
            "test_case_results": [asdict(tcr) for tcr in self.test_case_results],
        }

    def summary(self, public_only=False):
        if (not public_only and self.passed_all) or (public_only and self.passed_all_public):
            ret = f"{self.name} results: All test cases passed!"
            if (not public_only and self.passed_all) and \
                    any(tcr.test_case.success_message is not None for tcr in self.test_case_results):
                for tcr in self.test_case_results:
                    if tcr.test_case.success_message is not None:
                        ret += f"\n{tcr.test_case.name} message: {tcr.test_case.success_message}"
            return ret

        tcrs = self.test_case_results
        if public_only:
            tcrs = [tcr for tcr in tcrs if not tcr.test_case.hidden]

        tcr_summaries = []
        for tcr in tcrs:
            smry = ""
            if tcr.passed and tcr.test_case.success_message is not None:
                smry += f"{tcr.test_case.name} message: {tcr.test_case.success_message}\n\n"
            if not tcr.passed and tcr.test_case.failure_message is not None:
                smry += f"{tcr.test_case.name} message: {tcr.test_case.failure_message}\n\n"
            smry += f"{tcr.test_case.name} result:\n"
            smry += f"{indent(tcr.message.strip(), '    ')}\n\n"

            tcr_summaries.append(smry.strip())

        return f"{self.name} results:\n" + indent("\n\n".join(tcr_summaries), "    ")

    @classmethod
    @abstractmethod
    def from_file(cls, path):
        ...

    @abstractmethod
    def run(self, global_environment):
        ...
